#!/usr/bin/env python
# -*- coding: utf-8 -*-

#
# This file is part of the `plugy` python module
#
# Copyright
# 2018-2021
# EMBL & Heidelberg University
#
# Author(s): Dénes Türei (turei.denes@gmail.com)
#            Nicolas Peschke
#            Olga Ivanova
#
# Distributed under the GPLv3 License.
# See accompanying file LICENSE.txt or copy at
#     http://www.gnu.org/licenses/gpl-3.0.html
#
# Webpage: https://github.com/saezlab/plugy
#


import sys
import logging
import collections
import importlib as imp
import traceback
import typing
from dataclasses import dataclass, field

import numpy as np
import pandas as pd
import scipy.stats as stats
import statsmodels.stats.multitest as statsmod

import matplotlib as mpl
import matplotlib.pyplot as plt
import seaborn as sns

from .data.config import PlugyConfig
from .data.bd import ChannelMap, PlugSequence
from .data.pmt import PmtData
from .data.plug import PlugData
from . import misc as misc


module_logger = logging.getLogger(__name__)


@dataclass(init = False)
class PlugExperiment(object):
    config: PlugyConfig = field(default_factory = PlugyConfig)
    ignore_qc_result: bool = None
    run: bool = None
    init: bool = None
    plugs: bool = None
    samples: bool = None
    qc: bool = None
    analysis: bool = None
    font_scale: typing.Union[float, int] = 2
    scatter_dot_size: typing.Union[float, int] = 10


    def __init__(
            self,
            config: PlugyConfig = None,
            ignore_qc_result: bool = None,
            run: bool = None,
            init: bool = None,
            plugs: bool = None,
            qc: bool = None,
            analysis: bool = None,
            **kwargs
        ):

        self.config = config if config else PlugyConfig(**kwargs)

        for attr in (
            'run', 'init', 'plugs', 'qc', 'analysis', 'ignore_qc_result'
        ):

            from_call = locals()[attr]
            value = (
                getattr(self.config, attr)
                    if from_call is None else
                from_call
            )

            setattr(self, attr, value)


    def __post_init__(self):
        module_logger.info('Initializing PlugExperiment')
        module_logger.debug('Configuration:')
        for k, v in self.config.__dict__.items():
            module_logger.debug(f"{k}: {v}")

        self._set_workflow_param()

        if self.run:

            self.main()


    def reload(self):
        """
        Reloads the object from the module level.
        """

        modname = self.__class__.__module__
        mod = __import__(modname, fromlist = [modname.split('.')[0]])
        imp.reload(mod)
        new = getattr(mod, self.__class__.__name__)
        setattr(self, '__class__', new)


    def main(
            self,
            init: bool = None,
            plugs: bool = None,
            samples: bool = None,
            qc: bool = None,
            analysis: bool = None,
        ):

        steps = collections.OrderedDict()

        for attr in ('init', 'plugs', 'samples', 'qc', 'analysis'):

            call = locals()[attr]
            instance = getattr(self, attr)

            steps[attr] = call if call is not None else instance

        if not any(steps.values()):

            if self.config.has_samples_cycles:

                steps['qc'] = True
                steps['analysis'] = True

            elif self.config.has_barcode:

                steps['samples'] = True

            else:

                steps['plugs'] = True

        if steps['analysis'] or steps['qc']:

            steps['samples'] = True

        if steps['samples']:

            steps['plugs'] = True

        if steps['plugs']:

            steps['init'] = True

        module_logger.info('Workflow steps: %s' % misc.dict_str(steps))

        for step, enabled in steps.items():

            if enabled:

                module_logger.info('Executing step `%s`.' % step)
                getattr(self, '_%s' % step)()
                module_logger.info('Finished step `%s`.' % step)

            else:

                module_logger.info('Skipping step `%s`.' % step)


    def _init(self):

        self.seaborn_setup()
        self.load()


    def _plugs(self):

        self.detect_plugs()


    def _samples(self):

        self.detect_samples()


    def _qc(self):

        self.quality_control()


    def _analysis(self):

        self.drug_combination_analysis()
        self.close_figures()


    def _set_workflow_param(self):

        for attr in ('ignore_qc_result', 'run', 'init', 'plugs', 'samples'):

            here = getattr(self, attr)
            from_config = getattr(self.config, attr)

            setattr(
                self,
                attr,
                here if here is not None else from_config
            )


    def seaborn_setup(self, font_scale = None):


        sns.set_context(
            self.config.seaborn_context,
            font_scale = font_scale or self.font_scale,
            rc = self.config.seaborn_context_dict,
        )
        sns.set_style(
            self.config.seaborn_style,
            rc = self.config.seaborn_style_dict,
        )


    def load(self):

        self.load_channel_map()
        self.load_plug_sequence()
        self.load_pmt_data()


    def load_channel_map(self):

        self.channel_map = (

            ChannelMap(self.config.channel_file)

            if self.config.channel_file else

            None

        )


    def load_plug_sequence(self):

        self.plug_sequence = (

            PlugSequence.from_csv_file(
                self.config.seq_file,
                allow_lt4_valves = self.config.allow_lt4_valves,
            )

            if self.config.seq_file else

            None

        )


    def load_pmt_data(self):

        self.pmt_data = (

            PmtData(
                input_file = self.config.pmt_file,
                acquisition_rate = self.config.acquisition_rate,
                cut = self.config.cut,
                correct_acquisition_time = self.config.correct_acquisition_time,
                channels = self.config.channels,
                ignore_channels = self.config.ignore_channels,
                fake_gains = self.config.fake_gains,
                fake_gain_default = self.config.fake_gain_default,
                fake_gain_adaptive = self.config.fake_gain_adaptive,
                barcode_raw_threshold=self.config.barcode_raw_threshold,
                peak_min_threshold = self.config.peak_min_threshold,
                peak_max_threshold = self.config.peak_max_threshold,
                peak_min_distance = self.config.peak_min_distance,
                peak_min_prominence = self.config.peak_min_prominence,
                peak_max_prominence = self.config.peak_max_prominence,
                peak_min_width = self.config.peak_min_width,
                peak_max_width = self.config.peak_max_width,
                width_rel_height = self.config.width_rel_height,
                merge_peaks_distance = self.config.merge_peaks_distance,
                merge_peaks_by = self.config.merge_peaks_by,
                config = self.config,
            )

            if self.config.pmt_file else

            None

        )


    def detect_plugs(self):

        self.plug_data = PlugData(
            pmt_data = self.pmt_data,
            plug_sequence = self.plug_sequence,
            channel_map = self.channel_map,
            auto_detect_cycles = self.config.auto_detect_cycles,
            n_bc_adjacent_discards = self.config.n_bc_adjacent_discards,
            min_end_cycle_barcodes = self.config.min_end_cycle_barcodes,
            min_between_samples_barcodes =
                self.config.min_between_samples_barcodes,
            min_plugs_in_sample = self.config.min_plugs_in_sample,
            normalize_using_control = self.config.normalize_using_control,
            normalize_using_media_control_lin_reg =
                self.config.normalize_using_media_control_lin_reg,
            has_barcode = self.config.has_barcode,
            has_samples_cycles = self.config.has_samples_cycles,
            samples_per_cycle = self.config.samples_per_cycle,
            has_controls = self.config.has_controls,
            heatmap_second_scale = self.config.heatmap_second_scale,
            heatmap_override_scale = self.config.heatmap_override_scale,
            heatmap_override_second_scale =
                self.config.heatmap_override_second_scale,
            palette = self.config.palette,
            font_scale = self.font_scale,
            scatter_dot_size = self.scatter_dot_size,
            short_labels = self.config.short_labels,
            config = self.config,
        )


    def detect_samples(self):

        try:

            self.plug_data.detect_samples()

        except AssertionError:
            # In case labelling does not work because
            # the number of samples diverges from the expected number.
            # Plotting pmt data as a fallback

            module_logger.error(
                'Could not find any cycle with the expected number of '
                'samples. Potting the PMT data and exiting.'
            )

            self.plot_pmt_data()

            raise

        self.sample_data = self.get_sample_data()
        self.append_experiments()
        self.sample_statistics = self.calculate_statistics()


    def append_experiments(self, experiments: list = None):
        """
        Merges one or more other experiments by concatenating the plug and
        sample data frames. The time and cycle values will be modified in a
        way that the merged experiments follow each other without overlap.
        """

        experiments = (
            self.config.append
                if experiments is None else
            experiments
        )

        if not isinstance(experiments, (tuple, list)):

            experiments = tuple(experiments)

        for exp in experiments:

            if not isinstance(exp, tuple):

                exp = (exp,)

            self.plug_data.append(*exp)


    def plot_pmt_data(self, width = 300, **kwargs):

        qc_dir = self.ensure_qc_dir()

        pmt_overview_fig, pmt_overview_ax = plt.subplots(
            figsize = (width, 10)
        )
        kwargs['n_x_ticks'] = kwargs.get('n_x_ticks', width / 2)
        self.pmt_data.plot_pmt_data(
            pmt_overview_ax,
            **kwargs,
        )

        self.plug_data.highlight_plugs(
            axes = pmt_overview_ax,
            below_peak = False,
        )
        self.plug_data.highlight_samples(axes = pmt_overview_ax)
        self.plug_data.pmt_plot_add_thresholds(axes = pmt_overview_ax)

        if self.config.plot_git_caption:

            misc.add_git_hash_caption(pmt_overview_fig)

        pmt_overview_fig.tight_layout()
        png_path = qc_dir.joinpath(f"pmt_overview.png")
        pmt_overview_fig.savefig(png_path)
        plt.clf()

        module_logger.info(f"Plotted PMT data to {png_path}")


    def plot_plug_sequence(self):
        """
        Creates a compact plot of the detected peaks: x axis is the sequence
        of the peaks, y axis is the median intensity of the channels.
        """

        qc_dir = self.ensure_qc_dir()

        width = len(self.pmt_data.peak_df) / 20 + 3
        fig, ax = plt.subplots(figsize=(width, 3))
        self.pmt_data.plot_peak_sequence(ax)
        fig.tight_layout()

        path = qc_dir.joinpath(
            f"plug_sequence.{self.config.figure_export_file_type}"
        )
        fig.savefig(path)
        plt.clf()

        module_logger.info(f"Plotted plug sequence to {path}")


    def get_sample_data(self) -> pd.DataFrame:
        """
        Generates sample_data df that groups plugs for each sample and cycle calculating their median
        and how much their plug number diverges from the expected as produced by the BD using the PlugSequence
        :return: pd.DataFrame grouped by cycle_nr and sample_nr with the medians and additional plug count divergence
        """
        sample_data = self.plug_data.sample_df.groupby(["cycle_nr", "sample_nr"], as_index = False).median()
        divergence = self.get_plug_count_divergence()

        sample_data = sample_data.set_index(["cycle_nr", "sample_nr"]).join(divergence.set_index(["cycle_nr", "sample_nr"]))
        return sample_data


    def get_plug_count_divergence(self):
        """
        Compares generated and called number of plugs per sample and returns the difference
        :return: pd.Series with the divergence per sample
        """
        divergences = list()
        plug_sample_sequence = self.plug_sequence.get_samples(channel_map = self.channel_map).sequence
        for (cycle_nr, sample_nr), group in self.plug_data.sample_df.groupby(["cycle_nr", "sample_nr"]):
            divergence = (self.config.n_bc_adjacent_discards * 2) + len(group) - plug_sample_sequence[sample_nr].n_replicates
            divergences.append([cycle_nr, sample_nr, divergence])

        return pd.DataFrame(divergences, columns = ["cycle_nr", "sample_nr", "plug_count_divergence"])


    def get_contamination(self) -> pd.Series:
        """
        Normalizes filtered plug data to the means of the unfiltered data.
        :return: Series with the relative contamination of the plugs in sample_df
        """
        used_plugs = self.plug_data.plug_df.loc[self.plug_data.plug_df.cycle_nr.isin(self.plug_data.sample_df.cycle_nr.unique()) & self.plug_data.plug_df.barcode == 1]
        barcode_mean = used_plugs.loc[used_plugs.barcode].barcode_peak_median.mean()
        control_mean = self.plug_data.sample_df.control_peak_median.mean()
        norm_df = self.plug_data.sample_df.assign(
            norm_barcode = (
                self.plug_data.sample_df.barcode_peak_median /
                barcode_mean
            ),
            norm_control = (
                self.plug_data.sample_df.control_peak_median /
                control_mean
            )
        )

        return norm_df.norm_barcode / norm_df.norm_control


    def plot_plug_count_hist(self, axes: plt.Axes):
        """
        Plots the distribution of plug number divergence from the expected
        number by sample

        :param axes: plt.Axes object to draw on

        :return: plt.Axes object with the plot
        """

        axes = sns.countplot(
            x = self.sample_data.plug_count_divergence,
            ax = axes,
            color = self.config.palette[0],
        )
        axes.set_ylabel('Number of samples')
        axes.set_xlabel('Plug count deviation')

        return axes


    def quality_control(self):
        """
        Produces multiple QC plots and metrics to evaluate the technical
        quality of the PlugExperiment

        :return:
            True if quality is sufficient, False otherwise
        """

        self.seaborn_setup()

        qc_issues = []
        qc_dir = self.ensure_qc_dir()

        self.plot_medium_control_trends()

        self.plot_pmt_data()

        # Plotting plug numbers
        plug_count_hist_fig, plug_count_hist_ax = plt.subplots()
        plug_count_hist_ax = self.plot_plug_count_hist(
            axes = plug_count_hist_ax,
        )

        plug_count_hist_fig.tight_layout()
        if self.config.plot_git_caption:
            misc.add_git_hash_caption(plug_count_hist_fig)
        plug_count_hist_fig.savefig(
            qc_dir.joinpath(
                f"plug_count_hist.{self.config.figure_export_file_type}"
            )
        )
        plt.clf()

        self.plot_length_bias()

        # Plotting contamination
        contamination = self.get_contamination()
        if contamination.mean() > self.config.contamination_threshold:
            msg = (
                f"Contamination above threshold ({contamination.mean()} > "
                f"{self.config.contamination_threshold})"
            )
            module_logger.warning(msg)
            qc_issues.append(msg)

        contamination_hist_fig, contamination_hist_ax = plt.subplots(
            figsize = (8, 6.2),
        )
        sns.histplot(
            contamination,
            ax = contamination_hist_ax,
            color = self.config.palette[0],
            stat = 'density',
        )
        sns.kdeplot(
            contamination,
            ax = contamination_hist_ax,
            color = self.config.palette[1],
            linewidth = 2,
        )

        if self.config.figure_titles:

            contamination_hist_ax.set_title("Relative barcode contamination")

        contamination_hist_ax.set_xlabel(
            "Barcode control ratio\n"
            r"$\left[\frac{\overline{barcode_{median}}}"
            r"{\overline{control_{median}}}\right]$"
        )
        contamination_hist_ax.set_ylabel("Number of sample plugs")
        misc.add_git_hash_caption(contamination_hist_fig)
        contamination_hist_fig.tight_layout()
        contamination_hist_fig.savefig(
            qc_dir.joinpath(
                f"contamination_hist.{self.config.figure_export_file_type}"
            )
        )
        plt.clf()

        contamination_fig, contamination_ax = plt.subplots(
            2, 3,
            sharex = "all",
            sharey = "all",
            figsize = (18, 12),
        )

        for idx_y, channel in enumerate(
            ["readout_peak_median", "control_peak_median"]
        ):
            contamination_ax[idx_y][2] = self.plug_data.plot_contamination(
                channel_x = "barcode_peak_median",
                channel_y = channel,
                hue = "start_time",
                filtered = True,
                axes = contamination_ax[idx_y][2],
            )
            contamination_ax[idx_y][2].set_title("Filtered")
            contamination_ax[idx_y][2].get_legend().set_title('Time [s]')

            for idx_x, hue in enumerate(["start_time", "barcode"]):

                contamination_ax[idx_y][idx_x] = (
                    self.plug_data.plot_contamination(
                        channel_x = "barcode_peak_median",
                        channel_y = channel,
                        hue = hue,
                        filtered = False,
                        axes = contamination_ax[idx_y][idx_x],
                    )
                )

                contamination_ax[idx_y][idx_x].set_title("Unfiltered")

                leg_title = 'Time [s]' if hue == 'start_time' else 'Plug type'
                legend = contamination_ax[idx_y][idx_x].get_legend()
                legend.set_title(leg_title)

                if hue == 'barcode':
                    for t, l in zip(legend.texts, ['Sample', 'Barcode']):
                        t.set_text(l)

        if self.config.plot_git_caption:

            misc.add_git_hash_caption(contamination_fig)

        contamination_fig.tight_layout()
        contamination_fig.savefig(
            qc_dir.joinpath(
                f"contamination.{self.config.figure_export_file_type}"
            )
        )
        plt.clf()

        self.plot_control()
        self.plot_sample_cycles()

        self.length_grid()
        self.length_density()
        self.volume_density()
        self.sample_sd_violin()
        self.report_cv()

        plt.close('all')

        qc_successful = not qc_issues
        qc_fail_msg = (
            "Quality control failed due to the following reasons: "
            f"{', '.join(qc_issues)}. "
            "See also the QC plots for more information. "
            "In case you still want to continue, you can "
            "set the `ignore_qc_result` config parameter to True."
        )

        if qc_issues:
            module_logger.critical(qc_fail_msg)
        else:
            module_logger.info("Quality control successful")

        if not self.ignore_qc_result:

            assert qc_successful, qc_fail_msg

        return qc_successful


    def ensure_qc_dir(self):

        qc_dir = self.config.result_dir.joinpath("qc")

        if not qc_dir.exists():

            qc_dir.mkdir()

        return qc_dir


    def plot_medium_control_trends(self):

        if not self.plug_data.has_medium_control:

            module_logger.info(
                'Skipping medium control trends figure, looks like '
                'this experiment has no medium only samples.'
            )
            return

        qc_dir = self.ensure_qc_dir()

        # Plotting media control readout over experiment time
        media_control_fig, media_control_ax = plt.subplots(
            ncols = 2,
            figsize = (14, 7),
        )
        media_control_ax[0] = (
            self.plug_data.plot_medium_control_trends(
                axes = media_control_ax[0],
            )
        )
        media_control_ax[1] = (
            self.plug_data.plot_medium_control_trends(
                axes = media_control_ax[1],
                by_sample = True,
            )
        )

        media_control_fig.tight_layout()

        if self.config.plot_git_caption:
            misc.add_git_hash_caption(media_control_fig)

        path = qc_dir.joinpath(
            f"fs_medium_control.{self.config.figure_export_file_type}"
        )

        media_control_fig.savefig(path)
        plt.clf()

        module_logger.info('Medium control trends plotted to %s' % path)


    def plot_length_bias(self):
        """
        Creates a figure of readout signal vs. plug length by sample.
        """

        try:
            qc_dir = self.ensure_qc_dir()
            length_bias_plot = self.plug_data.plot_length_bias(col_wrap = 8)
            if self.config.plot_git_caption:
                misc.add_git_hash_caption(length_bias_plot.fig)
            length_bias_plot.fig.tight_layout()
            path = qc_dir.joinpath(
                f'length_bias.{self.config.figure_export_file_type}'
            )
            length_bias_plot.fig.savefig(path)
            plt.clf()
            module_logger.info('Plug length bias plotted to %s' % path)
        except:
            traceback.print_exc(file = sys.stdout)
            module_logger.error('Failed to plot length bias')


    def plot_control(self):
        """
        Creates a composite figure to investigate if the control channel
        shows a bias with other variables or correlation with the readout
        channel.
        """

        self.seaborn_setup(font_scale = self.config.font_scale * .7)

        qc_dir = self.ensure_qc_dir()

        # Plotting control
        control_fig = plt.figure(
            figsize = (20, 10),
            constrained_layout = False,
        )
        control_fig_gs = control_fig.add_gridspec(nrows = 2, ncols = 4)
        control_ax_sample_dist = control_fig.add_subplot(control_fig_gs[0, :])
        control_ax_control_regression = (
            control_fig.add_subplot(control_fig_gs[1, 0])
        )
        control_ax_cycle_dist = control_fig.add_subplot(control_fig_gs[1, 1])
        control_ax_readout_correlation = (
            control_fig.add_subplot(control_fig_gs[1, 2])
        )

        control_ax_control_regression = (
            self.plug_data.plot_control_regression(
                control_ax_control_regression
            )
        )
        control_ax_cycle_dist = self.plug_data.plot_control_cycle_dist(
            control_ax_cycle_dist
        )
        control_ax_sample_dist = self.plug_data.plot_control_sample_dist(
            control_ax_sample_dist
        )
        control_ax_readout_correlation = (
            self.plug_data.plot_control_readout_correlation(
                control_ax_readout_correlation
            )
        )

        grid = self.plug_data.plot_compound_heatmap(
            column_to_plot = 'control_peak_median',
        )

        control_fig.axes[-1] = grid.axes.flat[0]

        control_fig.tight_layout()
        if self.config.plot_git_caption:
            misc.add_git_hash_caption(control_fig)
        control_fig.savefig(
            qc_dir.joinpath(
                f"control_fluorescence.{self.config.figure_export_file_type}"
            )
        )
        plt.clf()

        self.seaborn_setup()


    def plot_sample_cycles(self, **kwargs):

        self.seaborn_setup(font_scale = self.font_scale * .7)

        # Plotting PMT overview
        try:
            qc_dir = self.ensure_qc_dir()
            sample_cycle_fig, sample_cycle_ax = (
                self.plug_data.plot_sample_cycles(**kwargs)
            )

            if self.config.plot_git_caption:

                misc.add_git_hash_caption(sample_cycle_fig)

            sample_cycle_fig.savefig(
                qc_dir.joinpath('sample_cycle_overview.png')
            )
            plt.clf()
        except:
            traceback.print_exc(file = sys.stdout)
            module_logger.error('Failed to plot sample cycles')

        self.seaborn_setup()


    def calculate_statistics(self) -> pd.DataFrame:
        """
        Calculates statistics
        """
        module_logger.info("Calculating statistics")
        media_data = self.plug_data.medium_only()
        compound_data = self.plug_data.sample_df[
            ~self.plug_data.sample_df.isin(media_data)
        ].dropna()

        group_columns = ["compound_a", "compound_b", "name"]
        sample_stats = compound_data.groupby(by = group_columns)[
            self.config.readout_analysis_column
        ].agg([np.mean, np.std])

        p_values = list()

        for combination, values in compound_data.groupby(by = group_columns):

            p_values.append(
                stats.ranksums(
                    x = values[self.config.readout_analysis_column],
                    y = media_data[self.config.readout_analysis_column]
                )[1]
            )

        sample_stats = sample_stats.assign(pval = p_values)
        significance, p_adjusted, _, alpha_corr_bonferroni = (
            statsmod.multipletests(
                pvals = sample_stats.reset_index().pval,
                alpha = self.config.alpha,
                method = "bonferroni",
            )
        )
        sample_stats = sample_stats.assign(
            p_adjusted = p_adjusted,
            significant = significance,
        )

        # Renaming columns to avoid shadowing mean function
        sample_stats.columns = [
            "mean_z_score",
            "std_z_score",
            "pval",
            "p_adjusted",
            "significant",
        ]

        # # Reindex based on sample_df from plug.PlugData object
        # sample_stats = sample_stats.reindex(self.plug_data.sample_df.name.unique())
        return sample_stats


    def drug_combination_analysis(self):
        """
        Analyzes drug combinations and produces result plots
        """
        module_logger.info("Running drug combination analysis")

        self.z_scores_violin_plot()
        self.z_scores_violin_plot(by_cycle = True)
        self.z_scores_heatmap()
        self.z_scores_heatmap(by_cycle = True)


    def z_scores_violin_plot(self, by_cycle: bool = False):

        # Overview violin plot with z-scores

        self.seaborn_setup(font_scale = self.font_scale * .7)

        cycles = (
            self.plug_data.sample_df.cycle_nr.unique()
                if by_cycle else
            [None]
        )

        fig, ax = plt.subplots(
            nrows = (
                self.plug_data.sample_df.cycle_nr.nunique()
                    if by_cycle else
                1
            ),
            figsize = (
                max(
                    4,
                    round(len(self.plug_data.sample_df.name.unique()) * 0.35)
                ),
                3 * (len(cycles) if by_cycle else 1) + 2
            ),
            sharex = True,
        )

        for i, cycle in enumerate(cycles):

            this_ax = ax if len(cycles) == 1 else ax.flat[i]

            self.plug_data.plot_compound_violins(
                ax = this_ax,
                column_to_plot = self.config.readout_analysis_column,
                cycle = cycle,
            )

        this_ax = ax if len(cycles) == 1 else ax.flat[0]
        # Getting y coordinates for asterisk from axis dimensions
        y_max = this_ax.axis()[3]

        # Labelling significant samples
        statistics = self.sample_statistics.reset_index()
        statistics = statistics.drop(["compound_a", "compound_b"], axis = 1)
        statistics = statistics.set_index("name")
        statistics = statistics.reindex(
            self.plug_data.sample_df.name.unique()
        )

        for idx, sample in enumerate(self.plug_data.sample_df.name.unique()):

            if sample != "Cell Control":

                if statistics.significant[idx]:

                    this_ax.annotate(
                        "*",
                        xy = (idx, y_max),
                        xycoords = "data",
                        textcoords = "data",
                        ha = "center",
                    )

        if not by_cycle:

            this_ax.set_title("Caspase activity z-scores", pad = 20)

        fig.tight_layout()

        if self.config.plot_git_caption:

            misc.add_git_hash_caption(fig)

        path = self.config.result_dir.joinpath(
            f"drug_comb_z_violins{'_by-cycle' if by_cycle else ''}."
            f"{self.config.figure_export_file_type}"
        )
        module_logger.info(f"Saving violin plots to {path}")
        fig.savefig(path)
        plt.clf()
        self.seaborn_setup()


    def z_scores_heatmap(self, by_cycle: bool = False):
        """
        Drug vs. drug heatmap showing the median readouts for each drug
        combination.
        """

        # Overview heatmap of z-scores

        # Labelling significant samples
        statistics = self.sample_statistics.reset_index()
        statistics = statistics.drop("name", axis = 1)
        statistics = statistics.set_index(["compound_a", "compound_b"])
        statistics = statistics.significant

        grid = self.plug_data.plot_compound_heatmap(
            column_to_plot = self.config.readout_analysis_column,
            annotation_df = statistics,
            by_cycle = by_cycle,
        )

        self._plot_base(
            grid,
            f"drug_comb_z_heatmap{'_by-cycle' if by_cycle else ''}",
            'drug combination heatmap',
            qc_dir = False,
        )


    def length_grid(self):
        """
        A pairs grid figure between plug lengths, fluorescent channels
        and the ratio of the readout and control channels.
        """

        grid = self.plug_data.length_grid()

        self._plot_base(grid, 'lengths_grid', 'plug length grid')


    def length_density(self):
        """
        Density plots and histograms about plug lengths. Three panels are
        created, with barcode plugs, sample plugs and both togther.
        """

        grid = self.plug_data.size_density()

        self._plot_base(grid, 'lengths_density', 'plug length density plots')


    def volume_density(self, flow_rate: float = 800.):
        """
        Density plots and histograms about plug volumes. Three panels are
        created, with barcode plugs, sample plugs and both togther.

        Args:
            flow_rate (float): The flow rate used at the data acquisition
                in microlitres per hour.
        """

        grid = self.plug_data.size_density(
            volume = True,
            flow_rate = flow_rate,
        )

        self._plot_base(grid, 'volumes_density', 'plug volume density plots')


    def sample_sd_violin(self):

        grid = self.plug_data.sample_sd_violin()

        self._plot_base(grid, 'sample_sd_violin', 'sample SD violin plots')


    def report_cv(self):

        cv = self.plug_data.mean_cv()

        for cycle, ccv in cv.items():

            for var, vcv in ccv.items():

                module_logger.info(
                    'Coefficient of variance, cycle %s, %s: '
                    '%.01f (%.01f-%.01f).' % (
                        str(cycle),
                        var,
                        vcv['mean'],
                        vcv['ci_low'],
                        vcv['ci_high'],
                    )
                )


    def _plot_base(
            self,
            fig,
            fname,
            log_msg = None,
            qc_dir = True,
        ):
        """
        Many plotting functions contain mostly repeated code. Here we attempt
        to contain these common parts in one function.

        Args:
            fig: A matplotlib Figure object, or some higher level object
                carrying a Figure object (such as a grid) in its `fig`
                attribute.
            fname (str): The file name, without extension.
            log_msg (str): How to call this figure in the log messages.
            qc_dir (bool): Save the plots in the `qc` subdirectory instead of
                directly into the `results` directory.
        """

        parent = None

        if not isinstance(fig, mpl.figure.Figure):

            if hasattr(fig, 'fig'):

                parent = fig
                fig = parent.fig

            else:

                raise ValueError(
                    'Don\'t know how to process a `%s` object.' % type(fig)
                )

        fig.tight_layout()

        if self.config.plot_git_caption:

            misc.add_git_hash_caption(fig)

        dir_path = self.ensure_qc_dir() if qc_dir else self.config.result_dir
        path = dir_path.joinpath(
            f"{fname}.{self.config.figure_export_file_type}"
        )

        log_msg = log_msg or fname
        module_logger.info(
            f"Saving {log_msg} figure to {path}"
        )

        save = parent.savefig if hasattr(parent, 'savefig') else fig.savefig

        save(path)

        plt.clf()


    def close_figures(self):

        plt.close('all')


    def pmt_data_to_excel(
            self,
            fname = None,
            sheet_name = 'PMT_data',
            **kwargs,
        ):

        self._to_excel(
            df = self.pmt_data.data,
            fname = fname,
            sheet_name = sheet_name,
            fname_suffix = 'pmt',
            label = 'PMT data',
            **kwargs,
        )


    def raw_plugs_to_excel(
            self,
            fname = None,
            sheet_name = 'Plugs_raw',
            **kwargs,
        ):

        self._to_excel(
            df = self.plug_data.plug_df,
            fname = fname,
            sheet_name = sheet_name,
            fname_suffix = 'plugs_raw',
            label = 'raw plug data',
            **kwargs,
        )


    def plugs_to_excel(self, fname = None, sheet_name = 'Plugs', **kwargs):

        self._to_excel(
            df = self.plug_data.sample_df,
            fname = fname,
            sheet_name = sheet_name,
            fname_suffix = 'plugs',
            label = 'plug data',
            **kwargs,
        )


    def samples_to_excel(
            self,
            fname = None,
            sheet_name = 'Samples',
            **kwargs,
        ):

        self._to_excel(
            df = self.sample_data,
            fname = fname,
            sheet_name = sheet_name,
            fname_suffix = 'samples',
            label = 'sample data',
            **kwargs,
        )


    def stats_to_excel(
            self,
            fname = None,
            sheet_name = 'Statistics',
            **kwargs,
        ):

        self._to_excel(
            df = self.sample_statistics,
            fname = fname,
            sheet_name = sheet_name,
            fname_suffix = 'stats',
            label = 'sample statistics',
            **kwargs,
        )


    def _to_excel(
            self,
            df,
            fname = None,
            fname_suffix = None,
            label = 'data',
            sheet_name = 'Sheet1',
            **kwargs,
        ):

        fname = (
            fname or
            '%s_%s.xlsx' % (self.pmt_data.input_file.stem, fname_suffix)
        )

        if 'index' not in kwargs:

            kwargs['index'] = isinstance(df.index, pd.MultiIndex)

        df.to_excel(fname, sheet_name = sheet_name, **kwargs)
        module_logger.info('Exported %s to %s' % (label, fname))


#!/usr/bin/env python
# -*- coding: utf-8 -*-

#
#  This file is part of the `plugy` python module
#
#  Copyright
#  2018-2020
#  EMBL, Heidelberg University
#
#  File author(s): Dénes Türei (turei.denes@gmail.com)
#                  Nicolas Peschke
#
#  Distributed under the GPLv3 License.
#  See accompanying file LICENSE.txt or copy at
#      http://www.gnu.org/licenses/gpl-3.0.html
#

import os

def _get_version():
    """
    Returns the version number.

    Reads file named ``__version__`` and returns it as a tuple of
    integers.

    :return:
        (*tuple*) -- Three element tuple containing the major, minor and
        micro version numbers.

    **Example:**
        >>> # If ``__version__`` contains ``1.11.0``:
        >>> _get_version()
        (1, 11, 0)
    """
    
    root_dir = os.path.abspath(os.path.dirname(__file__))
    
    with open(os.path.join(root_dir, '__version__'), 'r') as v:
        
        return tuple([int(i) for i in v.read().strip().split('.')])


_MAJOR, _MINOR, _MICRO = _get_version()
__version__ = '%d.%d.%d' % (_MAJOR, _MINOR, _MICRO)
__release__ = '%d.%d' % (_MAJOR, _MINOR)
__author__ = u'Dénes Türei & Nicolas Peschke'



#!/usr/bin/env python
# -*- coding: utf-8 -*-

#
# This file is part of the `plugy` python module
#
# Copyright
# 2018-2021
# EMBL & Heidelberg University
#
# Author(s): Dénes Türei (turei.denes@gmail.com)
#            Nicolas Peschke
#            Olga Ivanova
#
# Distributed under the GPLv3 License.
# See accompanying file LICENSE.txt or copy at
#     http://www.gnu.org/licenses/gpl-3.0.html
#
# Webpage: https://github.com/saezlab/plugy
#

import sys
import logging
import pickle
import importlib as imp
import warnings
import collections
import itertools
import functools
import inspect
import typing
from dataclasses import dataclass, field
import pathlib as pl

import tqdm

import pandas as pd
import numpy as np
import scipy.stats as stats
import skimage.filters

import matplotlib as mpl
import matplotlib.pyplot as plt
import matplotlib.patches as mpl_patch
import matplotlib.collections as mpl_coll
import seaborn as sns

from .. import misc
from ..data import pmt, bd
from ..data.config import PlugyConfig


module_logger = logging.getLogger(__name__)


@dataclass
class PlugData(object):

    pmt_data: pmt.PmtData
    plug_sequence: bd.PlugSequence
    channel_map: bd.ChannelMap
    auto_detect_cycles: bool = True
    merge_peaks_distance: float = 0.2
    n_bc_adjacent_discards: int = 1
    min_end_cycle_barcodes: int = 12
    min_between_samples_barcodes: int = 2
    min_plugs_in_sample: int = 1
    normalize_using_control: bool = False
    normalize_using_media_control_lin_reg: bool = False

    has_barcode: bool = True
    has_samples_cycles: bool = True
    samples_per_cycle: int = None
    has_controls: bool = True

    heatmap_second_scale: str = 'pos_ctrl'
    heatmap_override_scale: tuple = None
    heatmap_override_second_scale: tuple = None

    palette: tuple = None
    font_scale: typing.Union[float, int] = 2
    scatter_dot_size: typing.Union[float, int] = 10
    short_labels: bool = True

    config: PlugyConfig = field(default_factory = PlugyConfig)


    def __post_init__(self):

        module_logger.info(f"Creating PlugData object")
        module_logger.debug(
            f"Configuration: "
            f"{[f'{k}: {v}' for k, v in self.__dict__.items()]}"
        )

        self.detect_plugs()


    def reload(self):
        """
        Reloads the object from the module level.
        """

        modname = self.__class__.__module__
        mod = __import__(modname, fromlist = [modname.split('.')[0]])
        imp.reload(mod)
        new = getattr(mod, self.__class__.__name__)
        setattr(self, '__class__', new)


    def detect_plugs(self):
        """
        Finds plugs using the scipy.signal.find_peaks() method. Merges the
        plugs afterwards if merge_peaks_distance is > 0

        :return: DataFrame containing the plug data and a DataFrame
        containing information about the peaks as called by sig.find_peaks
        """
        module_logger.info("Finding plugs")
        self._detect_peaks()
        self._normalize_to_control()
        self._set_sample_param()
        self._set_barcoding()
        self._set_sample_cycle()


    def detect_samples(self):

        self._set_sample_param()
        self._count_samples_by_cycle()
        self._discard_cycles()
        self._label_samples()
        self._create_sample_df()
        self._add_z_scores()
        self._media_lin_reg_norm()
        self._check_sample_df_column(
            self.config.readout_analysis_column
        )
        self._check_sample_df_column(self.config.readout_column)
        self.calculate_z_factor()
        self.calculate_modified_z_factor()


    def _detect_peaks(self):
        """
        Detects peaks using scipy.signal.find_peaks().
        :return: Returns a DataFrame containing information about
            the peaks
        as called by sig.find_peaks
        """

        self.pmt_data.detect_peaks()
        self.plug_df = self.pmt_data.peak_df


    def _set_barcoding(self, **kwargs):


        def param_range(val):

            if isinstance(val, (tuple, list)):

                if len(val) == 2:

                    step = (val[1] - val[0]) / 20 * .9999999
                    val = tuple(val) + (step,)

                return np.arange(*val)

            else:

                return (val,)


        config = self.config
        param = config.barcoding_param_defaults.get(
            config.barcoding_method,
            {},
        )
        param.update(config.barcoding_param)
        param.update(kwargs)

        param = dict(
            (key, param_range(val))
            for key, val in param.items()
        )
        self._barcoding_param = collections.namedtuple(
            'BarcodeParam',
            sorted(param.keys())
        )
        self._barcode_result = collections.namedtuple(
            'BarcodeResult',
            [
                'sample_mismatch',
                'sample_freq_var',
            ]
        )
        self._barcode_eval = {}

        n_param = functools.reduce(
            lambda i, j: i * len(j),
            param.values(),
            1
        )
        pbar_desc = 'Adjusting barcode detection%s'

        with tqdm.tqdm(
            total = n_param,
            dynamic_ncols = True,
            desc = pbar_desc % '',
        ) as tq:

            for _values in itertools.product(*param.values()):

                _param = dict(zip(param.keys(), _values))
                self._barcoding_param_last = self._barcoding_param(
                    *(_param[f] for f in self._barcoding_param._fields)
                )
                tq.set_description(
                    pbar_desc % (
                        ' [%s]' % misc.ntuple_str(self._barcoding_param_last)
                    )
                )
                self._set_barcoding_base(**_param)
                self._evaluate_barcoding()
                tq.update()

        self._select_best_barcoding()
        self._set_barcoding_base(**self._barcode_best_param._asdict())

        self._sample_cycle_message()


    def _sample_cycle_message(self):

        if self.has_samples_cycles:

            module_logger.info(
                'Found %u cycles, %u with the expected number of samples. '
                'Sample count deviations: %s. '
                'Best barcode detection parameters: %s.' % (
                    len(self._sample_count_anomaly),
                    list(self._sample_count_anomaly.values()).count(0),
                    misc.dict_str(self._sample_count_anomaly),
                    misc.ntuple_str(self._barcode_best_param),
                )
            )


    def _set_barcoding_base(
            self,
            barcoding_method = None,
            **kwargs,
        ):
        """
        Creates a new boolean column `barcode` in the `plug_df` which is
        `True` if the plug is a barcode.
        """

        if not self.has_barcode:

            module_logger.debug(
                '`has_barcode` is False, skipping barcode identification.'
            )
            return

        config = self.config
        method_name = barcoding_method or config.barcoding_method
        method = '_set_barcoding_%s' % method_name
        param = config.barcoding_param_defaults.get(method_name, {})
        param.update(config.barcoding_param)
        param.update(kwargs)

        if hasattr(self, method):

            module_logger.debug('Barcode detection method: %s' % method)
            module_logger.debug(
                'Barcode detection parameters: %s' % misc.dict_str(param)
            )
            getattr(self, method)(**param)
            self._evaluate_barcoding()

        else:

            module_logger.error('No such method: `%s`' % method)


    def _set_barcoding_simple(self, times: float = None):

        self.plug_df = self.plug_df.assign(
            barcode = (
                self.plug_df.barcode_peak_median >
                self.plug_df.control_peak_median * times
            )
        )


    def _set_barcoding_adaptive(self, **kwargs):

        self._barcoding_thresholds = {}

        method_name = kwargs.pop('thresholding_method')
        method = getattr(skimage.filters, 'threshold_%s' % method_name)
        method_argnames = set(inspect.signature(method).parameters.keys())

        adaptive_method = (
            kwargs['adaptive_method']
                if 'adaptive_method' in kwargs else
            'simple'
        )

        channels = {}
        channel_names = {'barcode', 'control'}

        # adaptive thresholds on blue and orange
        for channel in channel_names:

            param = dict(
                (
                    key if key0 not in channel_names else key1,
                    val
                )
                for key, val, key0, key1 in
                (
                    ([key, val] + key.split('_', maxsplit = 1) + [None])[:4]
                    for key, val in kwargs.items()
                )
                if (
                    (
                        key in method_argnames and
                        '%s_%s' % (channel, key) not in kwargs
                    ) or (
                        key1 in method_argnames and
                        key0 == channel
                    )
                )
            )

            module_logger.debug(
                'Calling `%s` with parameters %s' % (
                    method_name,
                    misc.dict_str(param),
                )
            )

            channels[channel] = self.plug_df[
                '%s_peak_median' % channel
            ].to_numpy()
            shape = (1, channels[channel].shape[0])
            channels[channel].shape = shape

            threshold = method(channels[channel], **param)
            self._barcoding_thresholds[channel] = threshold.flatten()

        # setting the barcode based on the plugs' values and the blue or
        # the orange adaptive thresholds: either the blue is above threshold
        # or the orange is below the threshold
        if adaptive_method == 'simple':

            barcode_threshold =  self._barcoding_thresholds['barcode']
            control_threshold = self._barcoding_thresholds['control']
            self.plug_df['barcode'] = np.logical_or(
                channels['barcode'] > barcode_threshold,
                channels['control'] < control_threshold * .9,
            ).flatten()

        if adaptive_method in {'higher', 'slope'}:

            threshold_barcode_norm = (
                self._barcoding_thresholds['barcode'] /
                self._barcoding_thresholds['barcode'].max()
            )
            threshold_control_norm = (
                self._barcoding_thresholds['control'] /
                self._barcoding_thresholds['control'].max()
            )

        if adaptive_method == 'higher':

            factor = (
                kwargs['higher_threshold_factor']
                    if 'higher_threshold_factor' in kwargs else
                1.
            )
            self.plug_df['barcode'] = (
                threshold_barcode_norm >
                threshold_control_norm * factor
            ).flatten()

        if adaptive_method == 'slope':

            barcode_slope = np.diff(threshold_barcode_norm.flatten())
            barcode_slope = np.concatenate([0], barcode_slope)
            control_slope = np.diff(threshold_control_norm.flatten())
            control_slope = np.concatenate([0], control_slope)

        # adaptive threshold on blue:orange ratio
        param = dict(
            (key, val)
            for key, val in kwargs.items()
            if key in method_argnames
        )
        ratio = (
            self.plug_df.barcode_peak_median.to_numpy() /
            self.plug_df.control_peak_median.to_numpy()
        )
        shape = (1, max(ratio.shape))
        ratio.shape = shape
        ratio = ratio / ratio.max() * 10
        threshold = method(ratio, **param)
        self._barcoding_thresholds['ratio'] = threshold.flatten()
        self._barcoding_thresholds['_ratio'] = ratio.flatten()

        if adaptive_method == 'ratio':

            self.plug_df['barcode'] = (ratio > threshold).flatten()

        #param = self.config.adaptive_param.copy()
        #param.update(kwargs)

        #barcode_control = (
            #self.plug_df.barcode_peak_median /
            #self.plug_df.control_peak_median
        #).to_numpy()
        #barcode_control = self.plug_df.barcode_peak_median.to_numpy().copy()

        #shape = (1, barcode_control.shape[0])
        #barcode_control.shape = shape

        #threshold = skimage.filters.threshold_local(
            #barcode_control,
            #**param,
        #)

        #barcode = self.plug_df.barcode_peak_median > threshold.flatten()
        #barcode = barcode.to_numpy()
        #barcode.shape = shape

        #barcode = skimage.morphology.opening(
            #barcode,
            #selem = skimage.morphology.square(2),
        #)

        #self.plug_df['barcode'] = barcode.flatten()


    def _normalize_to_control(self):
        """
        Normalizes the readout channel by dividing its value by a control
        channel.
        """

        if self.normalize_using_control:

            self.plug_df = self.plug_df.assign(
                readout_per_control = (
                    self.plug_df.readout_peak_median /
                    self.plug_df.control_peak_median
                )
            )


    def quantify_interval(self, start_index, end_index):
        """
        Calculates median for each acquired channel between two data points

        :param start_index: Index of the first datapoint in pmt_data.data
        :param end_index: Index of the last datapoint in pmt_data.data

        :return: List with start_index end_index and the channels according
        to the order of config.channels
        """

        return self.pmt_data.quantify_interval(start_index, end_index)


    def _set_sample_cycle(self, debug = False):
        """
        Enumerates samples and cycles. Adds new columns to the
        :py:attr:`plug_df`: `cycle_nr`, `sample_nr` and `discard`.
        """

        if not self.has_samples_cycles:

            self.plug_df['cycle_nr'] = 0
            self.plug_df['sample_nr'] = 0
            self.plug_df['discard'] = False
            return

        # counters
        current_cycle = 0
        bc_peaks = 0
        sm_peaks = 0
        sample_in_cycle = -1
        # new vectors
        cycle = []
        sample = []
        discard = []
        # short names
        bc_bw_samples = self.min_between_samples_barcodes
        bc_adj_discards = self.n_bc_adjacent_discards
        bc_cycle_end = self.min_end_cycle_barcodes
        sm_min_len = self.min_plugs_in_sample

        if debug:

            print(
                'bc_bw_samples=%u, bc_adj_discards=%u '
                'bc_cycle_end=%u, sm_min_len=%u' % (
                    bc_bw_samples,
                    bc_adj_discards,
                    bc_cycle_end,
                    sm_min_len,
                )
            )

        for idx, bc in enumerate(self.plug_df.barcode):

            if debug:

                print(
                    'plug=%u, '
                    'current_cycle=%u, sample_in_cycle=%u, '
                    'sm_peaks=%u, bc_peaks=%u, barcode=%s' % (
                        idx,
                        current_cycle,
                        sample_in_cycle,
                        sm_peaks,
                        bc_peaks,
                        str(bc),
                    )
                )

            if bc:

                discard.append(True)
                cycle.append(current_cycle)
                sample.append(max(sample_in_cycle, 0))
                bc_peaks += 1

            else:

                # step the cycle counter
                # if this is the first plug of a sample
                if bc_peaks > 0 or sample_in_cycle < 0:

                    if (
                        (
                            bc_peaks >= bc_bw_samples and
                            sm_peaks >= sm_min_len
                        )
                        or
                        (
                            bc_peaks >= self.min_end_cycle_barcodes and
                            (current_cycle or sm_peaks)
                        )
                        or
                        sample_in_cycle < 0
                    ):

                        sample_in_cycle += 1

                        if (
                            bc_peaks >= self.min_end_cycle_barcodes and
                            # I hope this won't mess up the behaviour
                            # in any other scenario
                            # was like this before:
                            # sample_in_cycle > 0
                            # and possibly this will be better:
                            (
                                current_cycle or
                                sm_peaks
                            )
                        ):

                            current_cycle += 1
                            sample_in_cycle = 0

                        sm_peaks = 0

                    bc_peaks = 0

                sample.append(sample_in_cycle)
                cycle.append(current_cycle)

                # stepping the within sample plug counter
                sm_peaks += 1

                # Discarding barcode-adjacent plugs
                try:
                    if (
                        self.plug_df.barcode[idx - bc_adj_discards] or
                        self.plug_df.barcode[idx + bc_adj_discards]
                    ):
                        discard.append(True)
                    else:
                        discard.append(False)
                except KeyError:
                    discard.append(False)

        self.plug_df = self.plug_df.assign(
            cycle_nr = cycle,
            sample_nr = sample,
            discard = discard,
        )


    def _add_z_scores(self):
        # Calculating z-score on filtered data and inserting it after
        # readout_peak_median (index 5)

        if len(self.sample_df) > 1:

            self.sample_df.insert(
                loc = 5,
                column = "readout_peak_z_score",
                value = stats.zscore(self.sample_df.readout_peak_median),
            )

            if self.normalize_using_control:

                self.sample_df.insert(
                    loc = 6,
                    column = "readout_per_control_z_score",
                    value = stats.zscore(self.sample_df.readout_per_control),
                )

        else:

            module_logger.warning(
                f"Samples DataFrame contains {len(self.sample_df)} line(s), "
                f"omitting z-score calculation!"
            )


    def calculate_z_factor(self, modified = False):
        """
        Calculates the Z factor for each complete cycle according to
        https://en.wikipedia.org/wiki/Z-factor. Negative controls are
        the medium only samples and positive controls are the positive
        control samples with the mean of the corresponding single drug
        samples subtracted. Set the sample labels by the config options
        `medium_control_label` and `positive_control_label`. If the
        argument `modified` is True, it calculates a modified version
        of the Z factor, see `calculate_modified_z_factor`.
        """

        if not self.has_controls:

            return

        z_factors = []
        readout_col = self.readout_col

        for cycle in self.cycles:

            pos_control = self.positive_controls(
                subtract_single_drugs = True,
                cycle = cycle,
            )
            pos_control = pos_control[readout_col]

            medium_control = self.medium_only(cycle = cycle)
            medium_control = medium_control[readout_col]

            if modified:

                neg_control = self.negative_controls(cycle = cycle)
                neg_control = neg_control[readout_col]

                z_factor_numerator = 2 * (
                    np.std(neg_control) +
                    np.std(pos_control) +
                    np.std(medium_control)
                )
                z_factor_denominator = (
                    np.mean(pos_control) -
                    np.mean(medium_control) -
                    np.mean(neg_control)
                )

            else:

                z_factor_numerator = 3 * (
                    np.std(medium_control) +
                    np.std(pos_control)
                )
                z_factor_denominator = abs(
                    np.mean(pos_control) -
                    np.mean(medium_control)
                )

            z_factors.append(1 - (z_factor_numerator / z_factor_denominator))

        label = 'modified' if modified else ''
        attr = 'z_factors%s' % ('_%s' % label if label else '')

        setattr(self, attr, z_factors)

        z_factors = [round(e, 2) for e in z_factors]
        module_logger.info(
            f"Reporting {label} z-factor "
            f"by cycle: {z_factors}"
        )

        return z_factors


    def calculate_modified_z_factor(self):
        """
        Calculates a Z factor with the SD and mean of positive controls,
        medium controls and negative controls are all in the numerator and
        denominator, respectively. The positive + negative control samples
        are ignored instead of being part of the negative control group,
        while the positive control + medium samples are part of the positive
        control group. The means of the corresponding single drug samples
        are subtracted from the positive controls.
        """

        return self.calculate_z_factor(modified = True)


    @property
    def pos_ctrl_lab(self):
        """
        Returns the label of the positive control channel.
        """

        pos_ctrl_lab = misc.first(
            misc.to_set(self.config.positive_control_label) & self.channels
        )

        module_logger.debug(f'Positive control label: `{pos_ctrl_lab}`')

        return pos_ctrl_lab


    @property
    def neg_ctrl_lab(self):
        """
        Returns the label of the negative control channel.
        """

        neg_ctrl_lab = misc.first(
            misc.to_set(self.config.negative_control_label) & self.channels
        )

        module_logger.debug(f'Negative control label: `{neg_ctrl_lab}`')

        return neg_ctrl_lab


    @property
    def med_ctrl_lab(self):
        """
        Returns the label of the medium only channel in the current
        experiment.
        """

        med_ctrl_lab = misc.first(
            misc.to_set(self.config.medium_control_label) & self.channels
        )

        module_logger.debug(f'Medium control label: `{med_ctrl_lab}`')

        return med_ctrl_lab


    @property
    def readout_col(self):
        """
        Returns the column name for the readout column according to the
        current settings. The name of this column can vary because normally
        at each normalization or correction step we create a new column.
        """

        return self.config.readout_analysis_column


    @property
    def channels(self):
        """
        Returns a set of all channel labels in the current experiment.
        """

        return (
            set(self.sample_df.compound_a) |
            set(self.sample_df.compound_b)
        )


    @property
    def cycles(self):
        """
        Return the cycle identifiers in the current experiment.
        """

        return self.sample_df.cycle_nr.unique()


    def positive_controls(self, subtract_single_drugs = True, cycle = None):
        """
        Returns a subset of the `sample_df` with only the positive control
        samples.

        :param bool subtract_single_drugs: Subtract the means of the
            corresponding single drug samples from the positive control
            values.
        :param int cycle: Restrict the data to these cycles. If None, all
            cycles will be used.
        """

        if cycle is None:

            return pd.concat([
                self.positive_controls(
                    subtract_single_drugs = subtract_single_drugs,
                    cycle = cycle,
                )
                for cycle in self.cycles
            ])

        pos_ctrl_lab = self.pos_ctrl_lab
        neg_ctrl_lab = self.neg_ctrl_lab
        med_ctrl_lab = self.med_ctrl_lab
        readout_col = self.readout_col

        pos_control = self.sample_df.loc[
            (
                (
                    (self.sample_df.compound_b == pos_ctrl_lab) &
                    (self.sample_df.compound_a != neg_ctrl_lab)
                ) |
                (
                    (self.sample_df.compound_a == pos_ctrl_lab) &
                    (self.sample_df.compound_b != neg_ctrl_lab)
                )
            ) &
            (self.sample_df.cycle_nr == cycle)
        ]

        if subtract_single_drugs:

            single_drugs = self.sample_df.loc[
                (
                    (
                        (self.sample_df.compound_a == med_ctrl_lab) &
                        (self.sample_df.compound_b != pos_ctrl_lab) &
                        (self.sample_df.compound_b != neg_ctrl_lab)
                    ) |
                    (
                        (self.sample_df.compound_b == med_ctrl_lab) &
                        (self.sample_df.compound_a != pos_ctrl_lab) &
                        (self.sample_df.compound_a != neg_ctrl_lab)
                    )
                ) &
                (self.sample_df.cycle_nr == cycle)
            ]

            # these warnings stuff just filter out the pandas setting
            # with copy warnings, which are an annoying example of the
            # bad design of pandas
            with warnings.catch_warnings():

                warnings.simplefilter('ignore')

                single_drugs['compound'] = np.where(
                    single_drugs.compound_a == med_ctrl_lab,
                    single_drugs.compound_b,
                    single_drugs.compound_a
                )

            single_drugs = (
                single_drugs.groupby('compound')[readout_col].agg('mean')
            ).rename('single_drug_mean')

            with warnings.catch_warnings():

                warnings.simplefilter('ignore')

                pos_control['compound'] = np.where(
                    pos_control.compound_a == pos_ctrl_lab,
                    pos_control.compound_b,
                    pos_control.compound_a
                )

            pos_control = pos_control.join(
                single_drugs,
                on = 'compound'
            )
            pos_control[readout_col] = (
                pos_control[readout_col] -
                pos_control.single_drug_mean
            )

        return pos_control


    def negative_controls(self, cycle = None):
        """
        Returns a subset of the `sample_df` with only the negative control
        samples.

        :param int cycle: Restrict the data to these cycles. If None, all
            cycles will be used.
        """

        if cycle is None:

            return pd.concat([
                self.negative_controls(cycle = cycle)
                for cycle in self.cycles
            ])

        pos_ctrl_lab = self.pos_ctrl_lab
        neg_ctrl_lab = self.neg_ctrl_lab

        neg_control = self.sample_df.loc[
            (
                (self.sample_df.compound_a == neg_ctrl_lab) |
                (self.sample_df.compound_b == neg_ctrl_lab)
            ) &
            (
                (self.sample_df.compound_a != pos_ctrl_lab) &
                (self.sample_df.compound_b != pos_ctrl_lab)
            ) &
            (self.sample_df.cycle_nr == cycle)
        ]

        return neg_control


    def medium_only(self, cycle = None):
        """
        Returns a subset of the `sample_df` with the medium only samples.

        :param int cycle: Restrict the data to these cycles. If None, all
            cycles will be used.
        """

        if cycle is None:

            return pd.concat([
                self.medium_only(cycle = cycle)
                for cycle in self.cycles
            ])

        med_ctrl_lab = self.med_ctrl_lab

        medium_control = self.sample_df.loc[
            (self.sample_df.compound_b == med_ctrl_lab) &
            (self.sample_df.compound_a == med_ctrl_lab) &
            (self.sample_df.cycle_nr == cycle)
        ]

        return medium_control


    def get_media_control_lin_reg(self, readout_column: str = ""):
        """
        Calculates a linear regression over time of all media control plugs
        The readout_peak_median column is used to calculate the regression

        :return: Tuple with slope, intercept, rvalue, pvalue and stderr of
        the regression. See https://docs.scipy.org/doc/scipy/reference/
        generated/scipy.stats.linregress.html for more information about
        the returned values.
        """

        medium_control = self.medium_only()

        if not bool(len(medium_control)):

            module_logger.warning(
                'Could not find medium only control samples, unable to '
                'fit a linear regression on them. If the experiment '
                'contains such samples, check the config value of '
                '`medium_control_label` (currently `%s`).' %
                self.config.medium_control_label,
            )

            slope, intercept, rvalue, pvalue, stderr = (None,) * 5

        else:

            readout_column = readout_column or self.config.readout_column

            slope, intercept, rvalue, pvalue, stderr = stats.linregress(
                medium_control.start_time,
                medium_control[readout_column],
            )

        return slope, intercept, rvalue, pvalue, stderr


    def seaborn_setup(self):

        sns.set_context(
            self.config.seaborn_context,
            font_scale = self.config.font_scale,
            rc = self.config.seaborn_context_dict,
        )
        sns.set_style(
            self.config.seaborn_style,
            rc = self.config.seaborn_style_dict,
        )


    def plot_plug_pmt_data(
            self,
            axes: plt.Axes,
            cut: tuple = (None, None),
        ) -> plt.Axes:
        """
        Plots pmt data and superimposes rectangles with the called plugs upon
        the plot. This function has been replaced by
        ``plugy.data.pmt.PmtData.plot_pmt_data`` and is probably safe to
        remove.

        :param axes: plt.Axes object to plot to
        :param cut: tuple with (start_time, end_time) to subset the plot to
        a certain time range

        :return: plt.Axes object with the plot
        """

        axes = self.pmt_data.plot_pmt_data(axes, cut = cut)

        plug_df = self.plug_df
        sample_df = self.sample_df

        if cut[0] is not None:
            plug_df = plug_df.loc[plug_df.start_time >= cut[0]]
            sample_df = sample_df.loc[sample_df.start_time >= cut[0]]

        if cut[1] is not None:
            plug_df = plug_df.loc[plug_df.end_time <= cut[1]]
            sample_df = sample_df.loc[sample_df.end_time <= cut[1]]

        # Plotting light green rectangles that indicate
        # the used plug length and plug height
        bc_patches = list()
        readout_patches = list()

        for plug in plug_df.itertuples():

            if plug.barcode:

                bc_patches.append(
                    mpl_patch.Rectangle(
                        xy = (plug.start_time, 0),
                        width = plug.end_time - plug.start_time,
                        height = plug.barcode_peak_median,
                    )
                )


        for plug in sample_df.itertuples():

            readout_patches.append(
                mpl_patch.Rectangle(
                    xy = (plug.start_time, 0),
                    width = plug.end_time - plug.start_time,
                    height = plug.readout_peak_median,
                )
            )

        axes.add_collection(
            mpl_coll.PatchCollection(
                bc_patches,
                facecolors = self.config.colors['uv'],
                alpha = 0.4,
            )
        )
        axes.add_collection(
            mpl_coll.PatchCollection(
                readout_patches,
                facecolors = self.config.colors['green'],
                alpha = 0.4,
            )
        )

        return axes


    def highlight_plugs(self, axes: plt.Axes, below_peak: bool = False):

        if below_peak:

            return self.highlight_plugs_below_peak(axes)

        else:

            return self.highlight_plugs_vspan(axes)


    def highlight_plugs_below_peak(self, axes: plt.Axes):

        plug_patches = {
            'readout': [],
            'barcode': [],
        }

        xmin, xmax = axes.get_xlim()

        for plug in self.plug_df.itertuples():

            if plug.end_time < xmin or plug.start_time > xmax:

                continue

            plug_type = 'barcode' if plug.barcode else 'readout'
            xleft = max(plug.start_time, xmin)
            xright = min(plug.end_time, xmax)
            patch = mpl_patch.Rectangle(
                xy = (xleft, 0),
                width = xright - xleft,
                height = getattr(plug, '%s_peak_median' % plug_type),
            )
            plug_patches[plug_type].append(patch)

        colors = {
            'readout': self.config.colors['green'],
            'barcode': self.config.colors['blue'],
        }

        for key, color in colors.items():

            axes.add_collection(
                mpl_coll.PatchCollection(
                    plug_patches[key],
                    facecolors = color,
                    alpha = .4,
                )
            )

        return axes


    def highlight_plugs_vspan(self, axes: plt.Axes):

        xmin, xmax = axes.get_xlim()

        for plug in self.plug_df.itertuples():

            if plug.end_time < xmin or plug.start_time > xmax:

                continue

            color = self.config.colors['uv' if plug.barcode else 'green']

            module_logger.debug(
                'Highlighting plug between %.02f and %.02f (%s).' % (
                    max(plug.start_time, xmin),
                    min(plug.end_time, xmax),
                    color,
                )
            )

            axes.axvspan(
                xmin = max(plug.start_time, xmin),
                xmax = min(plug.end_time, xmax),
                facecolor = color,
                alpha = .4,
            )

        return axes


    def highlight_samples(self, axes: plt.Axes):

        if 'sample_nr' not in self.plug_df.columns:

            return axes

        samples = self.plug_df.groupby(
            ['cycle_nr', 'sample_nr']
        ).agg(
            {
                'start_time': 'min',
                'end_time': 'max',
            }
        )

        ymax = axes.get_ylim()[1]
        xmin, xmax = axes.get_xlim()

        for sample in samples.itertuples():

            if sample.start_time > xmax or sample.start_time < xmin:

                continue

            axes.axvspan(
                xmin = max(sample.start_time, xmin),
                xmax = min(sample.end_time, xmax),
                facecolor = '#AAAAAA33',
                edgecolor = 'black',
                linewidth = 1,
            )
            axes.text(
                x = max(sample.start_time, xmin),
                y = ymax - .15,
                s = '%u/%u' % (sample.Index[0] + 1, sample.Index[1] + 1),
                size = 'xx-large',
            )

        return axes


    def pmt_plot_add_thresholds(self, axes: plt.Axes):

        if not hasattr(self, '_barcoding_thresholds'):

            return

        time = (self.plug_df.start_time + self.plug_df.end_time) / 2

        for channel, threshold in self._barcoding_thresholds.items():

            if channel[0] == '_':

                continue

            color = self.config.channel_color(channel)

            sns.lineplot(
                x = time,
                y = threshold / threshold.max() * axes.get_ylim()[1] * .95,
                color = color,
                style = True,
                dashes = [(2,2)],
                linewidth = 1.,
                legend = False,
                ax = axes,
            )

        ratio = self._barcoding_thresholds['_ratio']
        ratio = ratio / ratio.max() * axes.get_ylim()[1] * .95

        sns.scatterplot(
            x = time,
            y = ratio,
            color = '#CC00CC',
            ax = axes,
        )


    def plot_cycle_pmt_data(self, axes: plt.Axes) -> plt.Axes:
        """
        Plots pmt data and superimposes filled rectangles for cycles with
        correct numbers of samples and
        unfilled rectangles for discarded cycles

        :param axes: plt.Axes object to plot to
        :return: plt.Axes object with the plot
        """

        module_logger.info("Plotting cycle data")
        axes = self.pmt_data.plot_pmt_data(axes = axes, cut = (None, None))

        used_cycle_patches = list()
        discarded_cycle_patches = list()
        patch_height = (
            self.pmt_data.data[["green", "orange", "uv"]].max().max()
        )

        for used_cycle in self.sample_df.groupby("cycle_nr"):

            cycle_start_time = used_cycle[1].start_time.min()
            cycle_end_time = used_cycle[1].end_time.max()
            used_cycle_patches.append(
                mpl_patch.Rectangle(
                    xy = (cycle_start_time, 0),
                    width = cycle_end_time - cycle_start_time,
                    height = patch_height,
                )
            )

        for detected_cycle in self.plug_df.groupby("cycle_nr"):

            cycle_start_time = detected_cycle[1].start_time.min()
            cycle_end_time = detected_cycle[1].end_time.max()
            discarded_cycle_patches.append(
                mpl_patch.Rectangle(
                    xy = (cycle_start_time, 0),
                    width = cycle_end_time - cycle_start_time,
                    height = patch_height,
                )
            )

            axes.text(
                x = (
                    (cycle_end_time - cycle_start_time) / 2 +
                    cycle_start_time
                ),
                y = 0.9 * patch_height,
                s = f"Cycle {detected_cycle[0]}",
                horizontalalignment = "center",
            )

        axes.add_collection(
            mpl_coll.PatchCollection(
                used_cycle_patches,
                facecolors = "green",
                alpha = 0.4,
            )
        )
        axes.add_collection(
            mpl_coll.PatchCollection(
                discarded_cycle_patches,
                edgecolors = "red",
                facecolors = "none",
                alpha = 0.4,
            )
        )

        return axes


    def _set_sample_param(self):
        """
        Sets up the parameters for counting and possibly labeling samples.
        """

        # Label samples in case channel map and plug sequence are provided
        self._has_sequence = (
            isinstance(self.channel_map, bd.ChannelMap) and
            isinstance(self.plug_sequence, bd.PlugSequence)
        )

        self.sample_sequence = (
            self.plug_sequence.get_samples(channel_map = self.channel_map)
                if self._has_sequence else
            None
        )
        self.expected_samples = (
            len(self.sample_sequence.sequence)
                if self._has_sequence else
            self.samples_per_cycle
        )


    def _ensure_sample_param(self):

        if not hasattr(self, 'expected_samples'):

            self._set_sample_param()


    def _evaluate_barcoding(self):

        self._ensure_sample_param()
        self._count_samples_by_cycle()
        self._evaluate_barcoding_base()


    def _evaluate_barcoding_base(self):

        if self.has_samples_cycles:

            sample_mismatch = sum(
                abs(a)
                for a in self._sample_count_anomaly.values()
            )
            sample_freq_var = 0

            for cycle_nr in self.plug_df.cycle_nr.unique():

                this_cycle = self.plug_df[self.plug_df.cycle_nr == cycle_nr]
                start_times = (
                    this_cycle.groupby('sample_nr').min('start_time')
                )
                sample_freq_var += start_times['start_time'].std()

            result = self._barcode_result(
                sample_mismatch = sample_mismatch,
                sample_freq_var = sample_freq_var,
            )

        else:

            result = (0, 0)

        self._barcode_eval[self._barcoding_param_last] = result


    def _select_best_barcoding(self):

        self._barcode_best_param = min(
            self._barcode_eval,
            key = self._barcode_eval.get,
        )


    def _adjust_sample_detection(self):
        """
        Labels sample_df with associated names and compounds according to the
        ChannelMap in the PlugSequence

        :param sample_df: pd.DataFrame with sample_nr column to associate
        names and compounds

        :return: pd.DataFrame with the added name, compound_a and b columns
        """

        if not self._has_sequence or self.config.adaptive:

            return

        blue_highest = self.config.blue_highest
        n_valid_cycles = collections.defaultdict(list)

        while True:

            self._count_samples_by_cycle()
            self._get_valid_cycles()
            n_valid_cycles[len(self.valid_cycles)].append(blue_highest)

            if (
                any(
                    0 < abs(d) < 4
                    for d in self._sample_count_anomaly.values()
                ) and
                blue_highest < 1.5
            ):

                blue_highest += .05
                self._set_barcode(blue_highest = blue_highest)
                self._set_sample_cycle()

            else:

                break

        blue_highest = n_valid_cycles[max(n_valid_cycles.keys())][0]
        self._set_barcode(blue_highest = blue_highest)
        self._set_sample_cycle()
        self._count_samples_by_cycle()

        module_logger.info(
            f"Adjusted `blue_highest` "
            f"to {blue_highest}."
        )


    def _count_samples_by_cycle(self, update = True):

        if not self.expected_samples:

            return

        if update:

            self._set_sample_cycle()

        self._samples_by_cycle = dict(
            (
                cycle_nr,
                cycle.sample_nr.nunique()
            )
            for cycle_nr, cycle in self.plug_df.groupby('cycle_nr')
        )

        self._sample_count_anomaly = dict(
            (
                cycle_nr,
                n_samples - self.expected_samples
            )
            for cycle_nr, n_samples in self._samples_by_cycle.items()
        )


    def _discard_cycles(self):

        for cycle_nr, cycle in self.plug_df.groupby('cycle_nr'):

            diff = self._sample_count_anomaly[cycle_nr]

            if diff:

                log_msg = (
                    f"Cycle {cycle_nr} detected between "
                    f"{cycle.start_time.min()}-{cycle.end_time.max()} "
                    f"contains {self._samples_by_cycle[cycle_nr]} samples, "
                    f"{'less' if diff < 0 else 'more'} than "
                    f"expected ({self.expected_samples})."
                )
                module_logger.info(log_msg)
                module_logger.info(f"Discarding cycle {cycle_nr}.")


                with warnings.catch_warnings():

                    warnings.simplefilter('ignore')
                    self.plug_df.discard[
                        self.plug_df.cycle_nr == cycle_nr
                    ] = True

        self._get_valid_cycles()
        self.n_cycles = len(self._samples_by_cycle)

        message = (
            self.pmt_data._detection_issues_message()
                if hasattr(self, 'pmt_data') else
            ''
        )

        if not self.valid_cycles:

            module_logger.critical(
                f"None of the {self.n_cycles} cycles is valid."
            )

            if message:

                module_logger.info(message)

        elif len(self.valid_cycles) != self.n_cycles:

            module_logger.info(
                f"Out of {self.n_cycles} only "
                f"{len(self.valid_cycles)} are valid."
            )

            if not self.auto_detect_cycles:

                module_logger.critical(
                    f"All cycles must have the expected number of samples."
                )
                self.valid_cycles = ()

            if message:

                module_logger.info(message)

            if not self.auto_detect_cycles:

                self.valid


    def _get_valid_cycles(self):

        self.valid_cycles = [
            cycle_nr
            for cycle_nr, diff in self._sample_count_anomaly.items()
            if not diff
        ]


    def _label_samples(self):

        assert self.valid_cycles, 'No valid cycles available.'

        module_logger.info('Labelling samples with compound names')

        df = self.plug_df
        seq = self.sample_sequence

        df['name'] = df.loc[~df.discard].sample_nr.apply(
            lambda nr: self.get_sample_name(nr, seq)
        )
        df['compound_a'] = df.loc[~df.discard].sample_nr.apply(
            lambda nr: self.get_compound_name(nr, 0)
        )
        df['compound_b'] = df.loc[~df.discard].sample_nr.apply(
            lambda nr: self.get_compound_name(nr, 1)
        )


    def _create_sample_df(self):

        sample_df = self.plug_df.loc[self.plug_df.discard == False]
        self.sample_df = sample_df.drop(columns = ['discard', 'barcode'])


    def get_compound_name(self, sample_nr, which):
        """
        :param sample_nr: Sample number in the sequence of samples.
        :param which: Either 0 or 1. Index within the compound combination.
        """

        seq = self.sample_sequence

        if self.channel_map is None:

            compound = 'Compound %u/%u' % (sample_nr, which)

        else:

            compound = self.channel_map.get_compounds(
                seq.sequence[sample_nr].open_valves
            )[which]

        return compound


    def get_sample_name(
            self,
            sample_nr: int,
            sample_sequence: bd.PlugSequence,
        ):
        """
        Returns a unified naming string for a sample.
        Concatenation of both compounds or single compound or cell control
        :param sample_nr: Sample number
        :param sample_sequence: bd.PlugSequence object to get open valves from
        """

        if self.channel_map is None:

            compounds = 'Sample %u' % sample_nr

        else:

            compounds = self.channel_map.get_compounds(
                sample_sequence.sequence[sample_nr].open_valves
            )
            compounds = [
                compound
                for compound in compounds
                if compound != "FS"
            ]
            compounds = ' + '.join(compounds) or 'Cell Control'

        return compounds


    def plot_sample_cycles(self, ylim: tuple = (None, None)):
        """
        Creates a plot with pmt data for the individual samples and cycles.
        :return: plt.Figure and plt.Axes object with the plot
        """

        names = self.sample_df.name.unique()
        cycles = sorted(self.sample_df.cycle_nr.unique())

        sample_cycle_fig, sample_cycle_ax = plt.subplots(
            nrows = len(names),
            ncols = len(cycles),
            figsize = (7 * len(cycles), 5 * len(names)),
            squeeze = False,
        )

        y_max = self.sample_df.readout_peak_median.max() * 1.1
        ylim = (
            ylim[0] or 0,
            ylim[1] or y_max
        )

        for idx_y, name in enumerate(names):

            for idx_x, cycle in enumerate(cycles):

                module_logger.debug(
                    f"Plotting sample {idx_y + 1} of {len(names)}, "
                    f"cycle {idx_x + 1} of {len(cycles)}"
                )
                sample_cycle_ax[idx_y][idx_x] = self.plot_sample(
                    name = name,
                    cycle_nr = cycle,
                    axes = sample_cycle_ax[idx_y][idx_x],
                )
                sample_cycle_ax[idx_y][idx_x].set_ylim(ylim)

        sample_cycle_fig.tight_layout()
        sample_cycle_fig.subplots_adjust(left = .15)

        return sample_cycle_fig, sample_cycle_ax


    def plot_sample(
            self,
            name: str,
            cycle_nr: int,
            axes: plt.Axes,
            offset: int = 10,
        ) -> plt.Axes:
        """
        Plots the PMT traces for a particular drug and cycle and

        :param name: Name of the drug combination/valve as listed in the
        PlugSequence
        :param cycle_nr: Number of the cycle
        :param axes: The plt.Axes object to draw on
        :param offset: How many seconds to plot left and right of the plugs

        :return: The plt.Axes object with the plot
        """

        peak_data = self.sample_df[
            (self.sample_df.cycle_nr == cycle_nr) &
            (self.sample_df.name == name)
        ]

        if len(peak_data) == 0:
            axes.text(0.5, 0.5, 'No Data')
        else:
            axes = self.plot_plug_pmt_data(
                axes = axes,
                cut = (
                    peak_data.start_time.min() - offset,
                    peak_data.end_time.max() + offset
                ),
            )

        axes.set_title(f'{name} • Cycle {cycle_nr + 1}', fontsize = 24)
        plt.setp(axes.get_xticklabels()[0], visible = False)

        return axes


    # QC Plots
    def plot_medium_control_trends(
            self,
            axes: plt.Axes,
            by_sample = False,
        ) -> plt.Axes:
        """
        Plots a scatter plot with readout medians for the media control
        over the experiment time.

        :param axes: plt.Axes object to draw on
        :param by_sample: True to plot swarmplot by sample number
        :return: plt.Axes object with the plot
        """

        self.seaborn_setup()

        if self.normalize_using_control:
            readout_column = 'readout_per_control'
            ylab = 'Readout control ratio'
        else:
            readout_column = 'readout_peak_median'
            ylab = 'Readout median'

        plot_data = self.medium_only()

        with warnings.catch_warnings():
            warnings.simplefilter('ignore')
            # stupid pandas again
            plot_data['cycle_nr'] = plot_data['cycle_nr'] + 1

        if by_sample:

            with warnings.catch_warnings():

                warnings.simplefilter('ignore')

                axes = sns.swarmplot(
                    x = 'sample_nr',
                    y = readout_column,
                    data = plot_data,
                    ax = axes,
                    hue = 'cycle_nr',
                    dodge = True,
                    palette = list(self.palette),
                    size = self.scatter_dot_size,
                )

            axes.set_xlabel('Sample')
            axes.get_legend().set_title('Cycle')

        else:

            slope, intercept, rvalue, _, _ = (
                self.get_media_control_lin_reg(readout_column)
            )

            with warnings.catch_warnings():

                warnings.simplefilter('ignore')

                axes = sns.scatterplot(
                    x = 'start_time',
                    y = readout_column,
                    data = plot_data,
                    ax = axes,
                    color = self.palette[0],
                    s = self.scatter_dot_size * 10,
                    alpha = .66,
                )

            misc.plot_line(slope, intercept, axes)

            label = axes.text(
                0.7,
                0.9,
                f'R²: {round(rvalue, 2)}',
                transform = axes.transAxes,
            )

            label.set_bbox(dict(facecolor = 'white', alpha = 0.8))
            axes.set_xlabel('Time [s]')

        if self.config.figure_titles:

            axes.set_title('Values from medium control plugs')

        axes.set_ylabel(ylab)

        return axes


    @property
    def has_medium_control(self):

        return bool(len(self.medium_only()))


    def add_length_column(self, df = 'plug'):
        """
        Creates a new column ``length`` in ``sample_df`` with the difference
        of plug start and end times.

        Args:
            df (str): Which data frame add the column to. Either "plug" or
                "sample".
        """

        attr = self._check_df(df, 'add_length_column')

        if attr:

            setattr(
                self,
                attr,
                self.lengths(getattr(self, attr)),
            )


    def add_volume_column(self, flow_rate: float = 800., df = 'plug'):
        """
        Creates a new column ``volume`` in ``sample_df`` with the volumes of
        the plugs in nanolitres.

        Args:
            flow_rate (float): The flow rate used at the data acquisition
                in microlitres per hour.
            df (str): Which data frame add the column to. Either "plug" or
                "sample".
        """

        attr = self._check_df(df, 'add_length_column')

        if attr:

            setattr(
                self,
                attr,
                self.volumes(getattr(self, attr), flow_rate = flow_rate),
            )


    def _check_df(self, df, fun):
        """
        Args:
            df (str): Which data frame: either "plug" or "sample".
            fun (str): Name of the calling function (to be shown in messages).
        """

        if df not in ('plug', 'sample'):

            module_logger.error(
                '`%s`: df must be either "plug" or "sample".' % fun
            )
            return

        attr = '%s_df' % df

        if not hasattr(self, attr):

            module_logger.error(
                '`%s`: %s data frame does not exist yet.' % (fun, attr)
            )
            return

        return attr


    def plot_length_bias(self, col_wrap: int = 3) -> sns.FacetGrid:
        """
        Plots each plugs fluorescence over its length grouped by valve.
        Also fits a linear regression to show if there is a correlation
        between the readout and the plug length indicating non ideal mixing.

        :param col_wrap: After how many subplots the column should be wrapped.
        :return: sns.FacetGrid object with the subplots
        """

        self.add_length_column(df = 'sample')

        df = self.sample_df

        df = df.groupby('name').filter(lambda x: x.shape[0] > 1)

        length_bias_plot = sns.lmplot(
            x = "length",
            y = "readout_peak_median",
            col = "name",
            data = df,
            col_wrap = col_wrap,
            scatter_kws = {
                'color': self.palette[0],
            },
            line_kws = {
                'color': self.palette[1],
            },
            height = 3.8,
            aspect = 1.1,
        )
        length_bias_plot.set_xlabels("Plug length [s]")
        length_bias_plot.set_ylabels("Readout [AU]")
        length_bias_plot.set_titles('{col_name}')

        length_bias_plot.set(
            ylim = (0, df.readout_peak_median.max()),
            xlim = (df.length.min(), df.length.max()),
        )

        return length_bias_plot


    def plot_contamination(
            self,
            channel_x: str,
            channel_y: str,
            axes: plt.Axes,
            filtered: bool = False,
            hue: str = "start_time",
            normalize: bool = False,
        ) -> plt.Axes:
        """
        Plots contamination as a scatter plot
        :param channel_x: Name of the channel to be plotted on the x axis
            (e.g. readout_peak_median, bc_peak_median, control_peak_median)
        :param channel_y: Name of the channel to be plotted on the y axis
            (e.g. readout_peak_median, bc_peak_median, control_peak_median)
        :param filtered: True if sample_df should be used, False if raw
            plug_df should be used
        :param axes: plt.Axes object to draw on
        :param hue: Name of the column in plug_df that is used to color the dots
        :param normalize: True plug data should be scaled by its mean
        :return: plt.Axes object with the plot
        """

        if filtered:
            norm_df = self.sample_df
        else:
            norm_df = self.plug_df.loc[
                self.plug_df.cycle_nr.isin(self.sample_df.cycle_nr.unique())
            ]

        palette = list(self.palette[:2]) if hue == 'barcode' else None

        if normalize:
            norm_df = norm_df.assign(
                norm_x = norm_df[channel_x] / norm_df[channel_x].mean(),
                norm_y = norm_df[channel_y] / norm_df[channel_y].mean(),
            )
            contamination_plot = sns.scatterplot(
                x = "norm_x",
                y = "norm_y",
                hue = hue,
                data = norm_df,
                ax = axes,
                alpha = .5,
                palette = palette,
            )

            axes.set_xlabel(
                '%s [%% of mean]' % (
                    channel_x.split('_')[0].capitalize()
                )
            )
            axes.set_ylabel(
                '%s [%% of mean]' % (
                    channel_y.split('_')[0].capitalize()
                )
            )

        else:
            contamination_plot = sns.scatterplot(
                x = channel_x,
                y = channel_y,
                hue = hue,
                data = norm_df,
                ax = axes,
                alpha = .5,
                palette = palette,
            )
            axes.set_xlabel(
                '%s [AU]' % channel_x.split('_')[0].capitalize()
            )
            axes.set_ylabel(
                '%s [AU]' % channel_y.split('_')[0].capitalize()
            )

        return contamination_plot


    def plot_control_regression(self, axes: plt.Axes) -> plt.Axes:
        """
        Plots a scatter plot of control peak medians over experiment time
            and applies a linear regression to it
        :param axes: plt.Axes object to draw on
        :return: plt.Axes object with the plot
        """
        axes = sns.regplot(
            x = "start_time",
            y = "control_peak_median",
            data = self.sample_df,
            ax = axes,
            color = self.palette[0],
            scatter_kws = {
                'alpha': .33,
            },
            line_kws = {
                'color': self.palette[1],
            },
        )

        if self.config.figure_titles:

            axes.set_title("Control: time bias")

        axes.set_ylabel("Control [AU]")
        axes.set_xlabel("Time [s]")

        return axes


    def plot_control_cycle_dist(self, axes: plt.Axes) -> plt.Axes:
        """
        Gathers control peak medians by cycle and plots a violin plot
        :param axes: plt.Axes object to draw on
        :return: plt.Axes object with the plot
        """

        axes = misc.seaborn_violin_fix(
            x = self.sample_df.cycle_nr + 1,
            y = self.sample_df.control_peak_median,
            ax = axes,
            color = self.palette[0],
            box_color = 'white',
            midpoint_color = self.palette[0],
            violin_border_width = 0,
        )

        if self.config.figure_titles:

            axes.set_title("Control: cycle bias")

        axes.set_ylabel("Control [AU]")
        axes.set_xlabel("Cycle")

        return axes


    @staticmethod
    def shorten_names(names):
        """
        Shortens the drug names to fit better on figures.
        """

        shorten = lambda name: ' + '.join(
            n[:5].strip()
            for n in name.split(' + ')
        )

        return [shorten(name) for name in names]


    def plot_control_sample_dist(self, axes: plt.Axes) -> plt.Axes:
        """
        Gathers control peak medians by sample and plots a violin plot
        :param axes: plt.Axes object to draw on
        :return: plt.Axes object with the plot
        """

        names = self.shorten_names(self.sample_df.name)

        axes = sns.violinplot(
            x = names,
            y = 'control_peak_median',
            data = self.sample_df,
            ax = axes,
            color = self.palette[0],
            linewidth = 0,
            width = .97,
        )

        axes = sns.boxplot(
            x = names,
            y = 'control_peak_median',
            data = self.sample_df,
            ax = axes,
            showbox = False,
            showfliers = False,
            showcaps = False,
            whiskerprops = {'linewidth': 0.0},
            medianprops = {'color': 'white'},
        )

        if self.config.figure_titles:

            axes.set_title('Control by sample')

        axes.set_ylabel('Control [AU]')
        axes.set_xlabel('Sample')

        for tick in axes.get_xticklabels():
            tick.set_rotation(90)

        return axes


    def plot_control_readout_correlation(self, axes: plt.Axes) -> plt.Axes:
        """
        Correlates control and readout peak medians
        :param axes: plt.Axes object to draw on
        :return: plt.Axes object with the plot
        """

        labels = {
            'sample_nr': 'Sample',
            'cycle_nr': 'Cycle',
        }

        axes = sns.scatterplot(
            x = 'control_peak_median',
            y = 'readout_peak_median',
            hue = 'sample_nr',
            style = self.sample_df.cycle_nr + 1,
            alpha = .7,
            data = self.sample_df,
            ax = axes,
        )

        if self.config.figure_titles:

            axes.set_title('Readout-control correlation')

        axes.set_xlabel('Control [AU]')
        axes.set_ylabel('Readout [AU]')
        lh, ll = axes.get_legend_handles_labels()
        axes.legend_.remove()
        ll = [labels[l] if l in labels else l for l in ll]
        axes.legend(
            lh,
            ll,
            handletextpad = .01,
            loc = 4,
            fancybox = False,
            borderpad = .2,
            edgecolor = '#FFFFFF00',
            framealpha = 1,
        )

        axes.set_ylim(
            self.sample_df.readout_peak_median.min() * 0.95,
            self.sample_df.readout_peak_median.max() * 1.05,
        )
        axes.set_xlim(
            self.sample_df.control_peak_median.min() * 0.95,
            self.sample_df.control_peak_median.max() * 1.5,
        )

        return axes


    def save(self, file_path: pl.Path):
        """
        Saves this PlugData object as pickle
        :param file_path: Path to the file to write to
        """
        with file_path.open("wb") as f:
            pickle.dump(self, f)


    def plot_compound_violins(
            self,
            ax: mpl.axes.Axes,
            column_to_plot: str = 'readout_peak_z_score',
            cycle: int = None,
        ) -> sns.FacetGrid:
        """
        Plots a violin plot per compound combination
        :param column_to_plot: Column to be plotted (e.g.
            readout_peak_z_score, readout_per_control_z_score).
        :param by_cycle: Produce separate plots for each cycle.
        :return: seaborn.FacetGrid object with the plot
        """

        self._check_sample_df_column(column_to_plot)

        data = self.sample_df

        if cycle is not None:

            data = data[data.cycle_nr == cycle]

        names = (
            self.shorten_names(data.name)
                if self.short_labels else
            data.name
        )

        misc.seaborn_violin_fix(
            x = names,
            y = column_to_plot,
            data = data,
            color = self.palette[0],
            width = 1.0,
            linewidth = 0,
            box_linewidth = 0.7,
            box_color = 'white',
            midpoint_color = self.palette[0],
            violin_border_width = 0,
            ax = ax,
        )

        ax.set_xticklabels(ax.get_xticklabels(), rotation = 90)
        ax.set_ylabel('Readout z-score')
        ax.set_xlim(-1, data.name.nunique())

        return ax


    def plot_compound_heatmap(
            self,
            column_to_plot: str,
            annotation_df: pd.DataFrame = None,
            annotation_column: str = "significant",
            by_cycle: bool = False,
            **kwargs
        ) -> sns.FacetGrid:
        """
        Plots a heatmap to visualize the different combinations
        :param column_to_plot: Name of the column to extract values from
        :param annotation_df: pd.DataFrame grouped by column, compound_a and
            compound_b for annotation
        :param annotation_column: Which column in annotation_df to use for
            the annotation
        :param by_cycle: Produce separate plots for each cycle.
        :return: seaborn.FacetGrid object with the plot
        """
        self._check_sample_df_column(column_to_plot)

        assert column_to_plot not in {"compound_a", "compound_b"},\
            f"You can not plot this coulumn on a heatmap: `{column_to_plot}`."

        if annotation_df is not None:

            annotation_df = annotation_df.reset_index()
            annotation_df = annotation_df.pivot(
                'compound_a',
                'compound_b',
                annotation_column,
            )
            annotation_df = (
                annotation_df.
                replace(True, '*').
                replace(False, '')
            )

        cycles = self.sample_df.cycle_nr.unique() if by_cycle else (None,)

        second_scale = (
            self.heatmap_second_scale in set(self.sample_df.compound_a) or
            self.heatmap_second_scale in set(self.sample_df.compound_b)
        )

        aspect_correction = 1.3 if second_scale else 1.0

        grid = sns.FacetGrid(
            data = self.sample_df,
            col = 'cycle_nr' if by_cycle else None,
            height = 7,
            aspect = (
                (.5 * len(cycles) if by_cycle else 1) *
                aspect_correction
            ),
        )

        for i, cycle in enumerate(cycles):

            data = self.sample_df
            data = data if cycle is None else data[data.cycle_nr == cycle]
            data = data[[column_to_plot, "compound_a", "compound_b"]]
            data = data.groupby(["compound_a", "compound_b"]).mean()
            data = data.reset_index()
            data = data.pivot("compound_a", "compound_b", column_to_plot)
            data = data.reindex(
                data.isna().
                sum(axis = 1).
                sort_values(ascending = False).
                index.
                to_list()
            )
            data = data[
                data.isna().
                sum(axis = 0).
                sort_values(ascending = True).
                index.
                to_list()
            ]

            if annotation_df is not None:

                annotation_df = (
                    annotation_df.reindex_like(data).
                    replace(np.nan, '')
                )

            if second_scale:

                second_scale_data = data.copy()
                data[self.heatmap_second_scale] = np.nan
                second_scale_data.loc[
                    :,
                    second_scale_data.columns != self.heatmap_second_scale
                ] = np.nan

                annot_second_scale = None

                if annotation_df is not None:

                    annot_second_scale = annotation_df.copy()
                    annot_second_scale.loc[
                        :,
                        annotation_df.columns != self.heatmap_second_scale
                    ] = ''

            ax = grid.axes.flat[i]

            vmin, vmax = self.heatmap_override_scale or (None, None)

            sns.heatmap(
                data,
                annot = annotation_df,
                fmt = "",
                ax = ax,
                vmin = vmin,
                vmax = vmax,
                **kwargs
            )

            if second_scale:

                vmin, vmax = (
                    self.heatmap_override_second_scale or
                    (None, None)
                )

                sns.heatmap(
                    second_scale_data,
                    annot = annot_second_scale,
                    cmap = 'viridis',
                    fmt = '',
                    ax = ax,
                    vmin = vmin,
                    vmax = vmax,
                    **kwargs
                )

            cycle_str = (' • Cycle %u' % (cycle + 1)) if by_cycle else ''
            unit_str = 'z-score' if 'z_score' in column_to_plot else '[AU]'
            var_str = column_to_plot.split('_')[0].capitalize()
            ax.set_title(
                f"{var_str} {unit_str}{cycle_str}"
            )
            ax.set_ylabel("")
            ax.set_xlabel("")

            if plt.matplotlib.__version__ == '3.1.1':

                ylim = ax.get_ylim()
                ax.set_ylim(ylim[0] + .5, ylim[1] - .5)

        return grid


    def _check_sample_df_column(self, column: str):
        """
        Checks if column is in sample_df
        :param column: Column to check
        """
        try:
            assert column in self.sample_df.columns.to_list(), f"Column {column} not in the column names of sample_df ({self.sample_df.columns.to_list()}), specify a column from the column names!"
        except AssertionError:
            module_logger.critical(f"Column {column} not in the column names of sample_df ({self.sample_df.columns.to_list()}), specify a column from the column names!")
            raise


    def _media_lin_reg_norm(self):
        """
        Normalizes sample_df using media control regression

        :return: updated sample_df
        """

        if self.normalize_using_media_control_lin_reg:

            sample_df = self.sample_df

            if self.normalize_using_control:
                readout_column = "readout_per_control"
            else:
                readout_column = "readout_peak_median"

            slope, intercept, _, _, _ = self.get_media_control_lin_reg(
                readout_column=readout_column
            )

            if slope and intercept:

                sample_df = sample_df.assign(
                    readout_media_norm = (
                        sample_df[readout_column] /
                        (sample_df["start_time"] * slope + intercept)
                    )
                )

                sample_df = sample_df.assign(
                    readout_media_norm_z_score = stats.zscore(
                        sample_df.readout_media_norm
                    )
                )

            else:

                module_logger.warning(
                    'Could not find medium only control samples, unable to '
                    'adjust readout values based on the drift of these '
                    'samples. If the experiment contains such samples, '
                    'check the config value of `medium_control_label` '
                    '(currently `%s`).' % self.config.medium_control_label,
                )

                new_readout_analysis_column = (
                    'readout_per_control_z_score'
                        if self.normalize_using_control else
                    'readout_z_score'
                )

                module_logger.critical(
                    'Using the column `%s` instead of `%s` in the analysis, '
                    'because the correction by medium only samples '
                    'failed.' % (
                        new_readout_analysis_column,
                        self.config.readout_analysis_column,
                    )
                )

                self.config.readout_analysis_column = (
                    new_readout_analysis_column
                )

            self.sample_df = sample_df


    def append(
            self,
            plug_df: pd.DataFrame,
            sample_df: pd.DataFrame = None,
            offset: float = 1.,
        ):

        # if PlugExperiment provided
        if hasattr(plug_df, 'plug_data'):

            plug_df = plug_df.plug_data

        # if PlugData provided
        if hasattr(plug_df, 'plug_df'):

            if sample_df is None and hasattr(plug_df, 'sample_df'):

                sample_df = plug_df.sample_df

            plug_df = plug_df.plug_df

        self.plug_df = self.concat_dfs(self.plug_df, plug_df, offset)

        if sample_df is not None and hasattr(self, 'sample_df'):

            self.sample_df = self.concat_dfs(
                self.sample_df,
                sample_df,
                offset,
            )

        if self.has_barcode:

            self._count_samples_by_cycle(update = False)
            self._sample_cycle_message()


    @staticmethod
    def concat_dfs(df1: pd.DataFrame, df2: pd.DataFrame, offset: float = 1.):

        tmax = df1.end_time.max()
        cmax = df1.cycle_nr.max()

        df2 = df2.copy()
        df2.start_time = df2.start_time + tmax + offset
        df2.end_time = df2.end_time + tmax + offset
        df2.cycle_nr = df2.cycle_nr - df2.cycle_nr.min() + cmax + 1

        return pd.concat([df1, df2]).reset_index()


    def length_grid(self) -> sns.PairGrid:
        """
        Creates a `seaborn.PairGrid` figure of plug lengths and readout
        values (https://seaborn.pydata.org/tutorial/distributions.html).

        Returns:
            (seaborn.PairGrid): A `PairGrid` object which can be saved
                into a matplotlib figure.
        """

        data = self.sample_and_barcode_plugs
        data = self.lengths(data)

        labels = {
            'length': 'Length [s]',
            'start_time': 'Time [s]',
            'barcode_peak_median': 'Barcode [AU]',
            'control_peak_median': 'Control [AU]',
            'readout_peak_median': 'Readout [AU]',
            'readout_per_control': 'R:C ratio',
            '': '',
        }

        legend_labels = {
            'barcode': '',
            'False': 'Sample plugs',
            'True': 'Barcode plugs',
            'cycle_nr': '',
            '0': 'Cycle 1',
            '1': 'Cycle 2',
            '2': 'Cycle 3',
            '3': 'Cycle 4',
        }

        grid = sns.PairGrid(
            data = data,
            hue = 'barcode',
            vars = [
                'length',
                'start_time',
                'barcode_peak_median',
                'control_peak_median',
                'readout_peak_median',
                'readout_per_control',
            ],
            palette = self.config.palette,
            despine = False,
        )

        grid.map_upper(sns.scatterplot, style = data.cycle_nr)
        grid.map_lower(sns.kdeplot, fill = True)
        grid.map_diag(sns.histplot, kde = False)

        for ax in grid.axes.flatten():

            # seaborn is crazy
            ax.get_yaxis().set_label_coords(-0.4,0.5)
            ax.yaxis.set_label_text(labels[ax.yaxis.get_label_text()])
            ax.xaxis.set_label_text(labels[ax.xaxis.get_label_text()])

        # seaborn is crazy
        grid.add_legend(title = '', bbox_to_anchor=(1.2, .5))

        # shameful seaborn
        for l in grid.fig.legends:

            for t in l.get_texts():

                if t.get_text() in legend_labels:

                    t.set_text(legend_labels[t.get_text()])

        return grid


    @property
    def sample_and_barcode_plugs(self):
        """
        Returns the plug data frame with the plugs marked for discard removed,
        and the sample and barcode plugs kept.
        """

        data = self.plug_df.loc[
            np.logical_not(self.plug_df.discard) |
            self.plug_df.barcode
        ]

        return data


    @staticmethod
    def lengths(df: pd.DataFrame) -> pd.DataFrame:
        """
        Creates a new column with plug or sample length (depending on the
        input data frame). The length is the difference of the start and end
        times in seconds.

        Args:
            df (pandas.DataFrame): A data frame with `start_time` and
                `end_time` columns.

        Returns:
            (pandas.DataFrame): A copy of the data frame with a new column
                called `length`.
        """

        df = df.copy()

        # these warnings stuff just filter out the pandas setting
        # with copy warnings, which are an annoying example of the
        # bad design of pandas
        with warnings.catch_warnings():

            warnings.simplefilter('ignore')
            df['length'] = df.end_time - df.start_time

        return df


    @classmethod
    def volumes(
            cls,
            df: pd.DataFrame,
            flow_rate: float = 800.,
        ) -> pd.DataFrame:
        """
        Creates a new column with plug volume. The volumes are calculated by
        the length of the plugs, if the data frame has no `length` column,
        it will be added too.

        Args:
            df (pandas.DataFrame): A data frame with `start_time` and
                `end_time` columns.
            flow_rate (float): The flow rate used at the data acquisition
                in microlitres per hour.

        Returns:
            (pandas.DataFrame): A copy of the data frame with a new column
                called `volume` containing the plug volumes in nanolitres.
                If the input data frame does not have a column `length`
                that will be added too.
        """

        df = df.copy()

        if 'length' not in df.columns:

            df = cls.lengths(df)

        # dividing by 3600 for hours -> seconds conversion
        # multiplying by 1000 for microlitres -> nanoliters conversion
        df['volume'] = df.length * flow_rate / 3.6

        return df


    def volume_stats(self, flow_rate: float = 800.):
        """
        Prints and returns statistics about the volume of the sample plugs.

        Args:
            flow_rate (float): The flow rate used at the data acquisition
                in microlitres per hour.

        Returns:
            (dict): A dictionary with cycle numbers and the string "all" as
                keys and dictionaries of statistics about the sample plug
                volumes as values.
        """

        data = self.volumes(self.sample_df, flow_rate = flow_rate)

        result = {
            'all': self._stats(data.volume)
        }

        for cycle in data.cycle_nr.unique():

            result[cycle] = self._stats(data[data.cycle_nr == cycle].volume)

        return result


    def size_density(
            self,
            volume: bool = False,
            flow_rate: float = 800.,
        ) -> sns.FacetGrid:
        """
        Density plot of plug sizes, either lengths in seconds or volumes
        in nanolitres.

        Args:
            volume (bool): Use plug volumes in nanolitres instead of lengths
                in seconds.
            flow_rate (float): The flow rate used at the data acquisition
                in microlitres per hour.

        Returns:
            (seaborn.FacetGrid): A grid with 3 subplots, each a density plot
                of plug sizes: barcode plugs, sample plugs and both
                together.
        """

        data = self.sample_and_barcode_plugs
        data = self.volumes(data, flow_rate = flow_rate)

        # these warnings stuff just filter out the pandas setting
        # with copy warnings, which are an annoying example of the
        # bad design of pandas
        with warnings.catch_warnings():

            warnings.simplefilter('ignore')

            data['group'] = [
                'Barcode' if bc else 'Sample'
                for bc in data.barcode
            ]

        data_both = data.copy()
        data_both['group'] = 'Both'
        data = pd.concat([data, data_both])

        grid = sns.FacetGrid(
            data,
            col = 'group',
            despine = False,
            height = 4,
        )
        grid.map(
            sns.histplot,
            'volume' if volume else 'length',
            kde = True,
            color = self.config.palette[0],
            line_kws = {'color': self.config.palette[1]},
        )

        for ax in grid.axes.flatten():

            # I can not believe seaborn has no better solution
            # for all these trivial tasks
            ax.set_title(
                ax.get_title().split('=')[-1].strip()
            )

        xlab = 'Plug %s [%s]' % (
            'volume' if volume else 'length',
            'nl' if volume else 's',
        )
        grid.axes.flatten()[0].xaxis.set_label_text('')
        grid.axes.flatten()[1].xaxis.set_label_text(xlab)
        grid.axes.flatten()[2].xaxis.set_label_text('')

        return grid


    def sample_sd_df(self) -> pd.DataFrame:
        """
        Creates a data frame of readout standard deviations and coefficients
        of variation (CV).
        """

        cv = lambda x: np.std(x) / np.mean(x) * 100.
        cv_stat = {'label': 'cv', 'fun': cv}

        return self.sample_stats(
            readout_peak_median = ('std', cv_stat),
            readout_per_control = ('std', cv_stat),
            readout_media_norm = ('std', cv_stat),
        )


    def sample_stats(self, long: bool = True, **kwargs) -> pd.DataFrame:
        """
        Calculates statistics for each sample in the samples data frame.

        Args:
            long (bool): Return a long or wide format data frame.
            kwargs: A mapping where keys are column names while values are
                aggregate function names or functions. For one column more
                than one statistics can be calculated by passing a tuple
                as value.

        Returns:
            (pandas.DataFrame): A data frame with each row representing
                one statistics for one sample, or each row representing
                one sample with all its statistics, if `long` is False.
        """

        agg_args = dict(
            (
                '%s___%s' % (
                    colname,
                    (
                        fun
                            if isinstance(fun, str) else
                        fun.__name__
                            if (
                                hasattr(fun, '__name__') and
                                fun.__name__ != '<lambda>'
                            ) else
                        fun['label']
                            if (
                                isinstance(fun, dict) and
                                'label' in fun
                            ) else
                        'agg%u' % i
                    )
                ),
                pd.NamedAgg(
                    colname,
                    fun['fun']
                        if isinstance(fun, dict) and 'fun' in fun else
                    fun,
                )
            )
            for colname, funs in kwargs.items()
            for i, fun in enumerate(
                funs
                    if isinstance(funs, tuple) else
                (funs,)
            )
            if colname in self.sample_df.columns
        )

        data = (
            self.sample_df.
            groupby(['cycle_nr', 'sample_nr']).
            agg(**agg_args)
        )

        if long:

            data = data.reset_index().melt(
                id_vars = ['cycle_nr', 'sample_nr'],
                var_name = 'var_stat',
                value_name = 'value',
            )

            data['stat'] = [v.split('___')[-1] for v in data['var_stat']]
            data['var'] = [v.split('___')[0] for v in data['var_stat']]

        return data


    def mean_cv(self):
        """
        Calculates the coefficient of variation for all cycles and the
        whole screen, for all available readout metrics.
        """

        cv_data = self.sample_sd_df()
        cv_data = cv_data[cv_data['stat'] == 'cv']

        readout_vars = cv_data['var'].unique()
        cycles = tuple(cv_data['cycle_nr'].unique()) + ('all',)

        result = collections.defaultdict(dict)

        for cycle, var in itertools.product(cycles, readout_vars):

            cv = cv_data[cv_data['var'] == var]

            if cycle != 'all':

                cv = cv[cv['cycle_nr'] == cycle]

            cv = cv.value

            result[cycle][var] = {
                'mean': cv.mean(),
                'ci_low': cv.mean() - cv.std(),
                'ci_high': cv.mean() + cv.std(),
            }

        return dict(result)


    def sample_sd_violin(self) -> sns.FacetGrid:

        def seaborn_violin_fix_fix(*args, **kwargs):
            """
            A tribute to seaborn's genious design idea of passing
            map arguments only as positional, not as keyword args.
            Congratulations!
            """

            x, y = args

            misc.seaborn_violin_fix(
                x = x,
                y = y,
                **kwargs
            )


        labels = {
            'readout_peak_median': 'Readout\n',
            'readout_per_control': 'Readout:control\nratio',
            'readout_media_norm': 'R:C corrected by\nmedium control',
        }

        data = self.sample_sd_df()

        data.cycle_nr = data.cycle_nr + 1

        grid = sns.FacetGrid(
            data,
            col = 'var',
            despine = False,
            height = 5,
        )

        bg_color = self.config.palette[0]
        fg_color = 'white'

        grid.map(
            seaborn_violin_fix_fix,
            # seaborn, is this serious??
            'cycle_nr',
            'value',
            color = bg_color,
            midpoint_color = bg_color,
            box_color = fg_color,
            violin_border_width = 0,
        )

        for ax in grid.axes.flatten():

            # I can not believe seaborn has no better solution
            # for all these trivial tasks
            ax.set_title(
                labels[ax.get_title().split('=')[-1].strip()]
            )

        grid.axes.flatten()[0].yaxis.set_label_text('Standard deviation')
        grid.axes.flatten()[0].xaxis.set_label_text('')
        grid.axes.flatten()[1].xaxis.set_label_text('Cycle')

        if len(grid.axes.flatten()) > 2:

            grid.axes.flatten()[2].xaxis.set_label_text('')

        return grid


    @staticmethod
    def _stats(data: pd.Series) -> dict:
        """
        For a numeric `Series` creates a dictionary of basic descriptive
        statistics, such as mean, median, quantiles, standard deviation.

        Args:
            (pandas.Series): A pandas.Series or numpy.ndarray object
                of numeric data type.

        Returns:
            (dict): A dictionary of descriptive statistics, with strings
                as keys and numbers as values.
        """

        return {
            'mean': data.mean(),
            'median': data.median(),
            'q1': data.quantile(.25),
            'q2': data.quantile(.75),
            'min': data.min(),
            'max': data.max(),
            'ci90_low': data.mean() - 1.645 * data.std(),
            'ci90_high': data.mean() + 1.645 * data.std(),
            'sd': data.std(),
        }


    def sample_sd(self, variable = 'readout_peak_median', silent = False):
        """
        Prints and returns the standard deviation of a variable within
        samples, with confidence intervals.

        Args:
            variable (str): The name of the variable to calculate the
                standard deviations from.

        Returns:
            (dict): A dictionary with cycle numbers and the string "all" as
                keys and dictionaries of statistics about the standard
                deviations of samples as values.
        """

        def sd_stats(data):

            return self._stats(data.value)


        args = {variable: 'std'}
        data = self.sample_stats(**args)
        result = {}

        for cycle in data.cycle_nr.unique():

            result[cycle] = sd_stats(data[data.cycle_nr == cycle])

        result['all'] = sd_stats(data)

        if not silent:

            sys.stdout.write(
                '\nStandard deviation of `%s` within samples:\n\n' % variable
            )
            sys.stdout.write(
                pd.DataFrame(result).__repr__()
            )
            sys.stdout.write('\n\n')
            sys.stdout.flush()

        return result



#!/usr/bin/env python
# -*- coding: utf-8 -*-

#
# This file is part of the `plugy` python module
#
# Copyright
# 2018-2021
# EMBL & Heidelberg University
#
# Author(s): Dénes Türei (turei.denes@gmail.com)
#            Nicolas Peschke
#            Olga Ivanova
#
# Distributed under the GPLv3 License.
# See accompanying file LICENSE.txt or copy at
#     http://www.gnu.org/licenses/gpl-3.0.html
#
# Webpage: https://github.com/saezlab/plugy
#


import logging
import tempfile
import unittest
import unittest.mock

import pathlib as pl

import numpy as np
import pandas as pd
import pandas._testing as pd_test
import scipy.signal as sig
import scipy.ndimage.filters as fil

from ..data import plug
from ..data import pmt
from ..data import bd
from ..data import config

import matplotlib.pyplot as plt

logging.basicConfig(
    level = logging.DEBUG,
    format = '%(asctime)s %(name)-12s %(levelname)-8s %(message)s',
    datefmt = '%d.%m.%y %H:%M:%S',
)

# disable annoying pandas SettingWithCopy warnings
pd.options.mode.chained_assignment = None


class TestPlugData(unittest.TestCase):


    def setUp(self) -> None:

        self.clean_data = pd.DataFrame()
        self.noisy_data = pd.DataFrame()

        self.signal_length = 8
        self.acquisition_rate = 300
        self.filter_size = self.acquisition_rate / 6
        self.seed = 0
        self.noise_sigma = 0.13
        self.pseudocount = 0.00001

        # Get precise simulated experiment time
        self.time = np.linspace(
            0,
            self.signal_length,
            int(self.signal_length * self.acquisition_rate)
        )

        self.clean_data = self.clean_data.assign(time = self.time)

        # Create clean square wave with period 2 pi and
        # with its first rising edge at pi
        self.clean_data = self.clean_data.assign(
            green = (
                (sig.square(
                    2 * np.pi * 0.5 * (self.clean_data.time - 1)
                )) +
                1
            ) / 2
        )
        self.clean_data = self.clean_data.assign(
            uv = self.clean_data.green
        )
        self.clean_data = self.clean_data.assign(
            orange = self.clean_data.green
        )

        self.clean_data.loc[self.clean_data.time > 4, "orange"] = 0
        self.clean_data.loc[
            (self.clean_data.time < 4) | (self.clean_data.time > 8),
            "uv"
        ] = 0
        self.clean_data.loc[
            (self.clean_data.time < 3) | (self.clean_data.time > 4),
            "green"
        ] = 0

        repeats = 5
        for _ in range(repeats):
            try:
                # noinspection PyUnboundLocalVariable
                tmp_data = tmp_data.append(self.clean_data) # noqa: F821
            except UnboundLocalError:
                tmp_data = self.clean_data

        # noinspection PyUnboundLocalVariable
        self.clean_data = tmp_data.reset_index(drop = True)

        self.clean_data = self.clean_data.assign(
            time = np.linspace(
                0,
                self.signal_length * repeats,
                int(self.signal_length * repeats * self.acquisition_rate)
            )
        )
        self.clean_data = pd.concat(
            [self.clean_data] +
            [self.clean_data.tail(1)] *
            100
        )
        self.clean_data.iloc[-100:].time = np.linspace(
            max(self.clean_data.time),
            max(self.clean_data.time) + .33,
            100
        )
        self.clean_data.reset_index()

        # End of cycle
        self.clean_data.loc[
            (self.clean_data.time > 13) & (self.clean_data.time < 24),
            "orange"
        ] = 0
        self.clean_data.loc[
            (self.clean_data.time > 13) & (self.clean_data.time < 24),
            "green"
        ] = 0

        self.clean_data.loc[
            (self.clean_data.time > 13) & (self.clean_data.time < 14),
            "uv"
        ] = 1
        self.clean_data.loc[
            (self.clean_data.time > 15) & (self.clean_data.time < 16),
            "uv"
        ] = 1
        self.clean_data.loc[
            (self.clean_data.time > 17) & (self.clean_data.time < 18),
            "uv"
        ] = 1

        self.clean_data = self.clean_data.assign(
            green = self.clean_data.green * 0.9
        )
        self.clean_data = self.clean_data.assign(
            orange = self.clean_data.orange * 0.8
        )

        self.clean_data = self.clean_data.assign(
            green = self.clean_data.green + self.pseudocount,
            uv = self.clean_data.uv + self.pseudocount,
            orange = self.clean_data.orange + self.pseudocount
        )

        # Filter the clean signal with a mean filter
        # to get slightly rounded edges
        self.noisy_data = self.noisy_data.assign(time = self.clean_data.time)
        self.noisy_data = self.noisy_data.assign(
            green = fil.convolve1d(
                input = self.clean_data.green,
                weights = np.array(np.repeat(1, self.filter_size))
            ) /
            self.filter_size
        )
        self.noisy_data = self.noisy_data.assign(
            uv = fil.convolve1d(
                input = self.clean_data.uv,
                weights = np.array(np.repeat(1, self.filter_size))
            ) /
            self.filter_size
        )
        self.noisy_data = self.noisy_data.assign(
            orange = fil.convolve1d(
                input = self.clean_data.orange,
                weights = np.array(np.repeat(1, self.filter_size))
            ) /
            self.filter_size
        )

        # Add gaussian noise to the clean square wave
        np.random.seed(self.seed)

        self.noisy_data = self.noisy_data.assign(
            green =
                self.noisy_data.green +
                np.random.normal(
                    scale = self.noise_sigma,
                    size = len(self.noisy_data.green)
                )
        )

        np.random.seed(self.seed)

        self.noisy_data = self.noisy_data.assign(
            uv =
                self.noisy_data.uv +
                np.random.normal(
                    scale = self.noise_sigma,
                    size = len(self.noisy_data.uv)
                )
        )

        np.random.seed(self.seed)

        self.noisy_data = self.noisy_data.assign(
            orange =
                self.noisy_data.orange +
                np.random.normal(
                    scale = self.noise_sigma,
                    size = len(self.noisy_data.orange)
                )
        )

        self.cycle_data = pd.DataFrame({
            "start_time": [
                1.0, 3.0, 5.0, 7.0, 9.0, 11.0, 13.0, 15.0, 17.0, 21.0,
                23.0, 25.0, 27.0, 29.0, 31.0, 33.0, 35.0, 37.0, 39.0,
            ],
            "end_time":  [
                2.0, 4.0, 6.0, 8.0, 10.0, 12.0, 14.0, 16.0, 18.0, 22.0,
                24.0, 26.0, 28.0, 30.0, 32.0, 34.0, 36.0, 38.0, 40.0,
            ],
            "barcode_peak_median": [
                0.0, 0.0, 1.0, 1.0,  0.0, 0.0, 1.0, 1.0, 1.0, 1.0,
                1.0, 0.0, 0.0, 1.0, 1.0, 0.0, 0.0, 1.0, 1.0,
            ],
            "control_peak_median": [
                0.8, 0.8, 0.0, 0.0, 0.8, 0.8, 0.0, 0.0, 0.0, 0.0,
                0.0, 0.8, 0.8, 0.0, 0.0, 0.8, 0.8, 0.0, 0.0,
            ],
            "readout_peak_median": [
                0.0, 0.9, 0.0, 0.0, 0.0, 0.9, 0.0, 0.0, 0.0, 0.0,
                0.0, 0.0, 0.9, 0.0, 0.0, 0.0, 0.9, 0.0, 0.0,
            ],
            "barcode": [
                False, False, True, True, False, False,
                True, True, True, True, True, False, False,
                True, True, False, False, True, True,
            ]
        })

        self.cycle_data = self.cycle_data.assign(
            barcode_peak_median =
                self.cycle_data.barcode_peak_median +
                self.pseudocount,
            control_peak_median =
                self.cycle_data.control_peak_median +
                self.pseudocount,
            readout_peak_median =
                self.cycle_data.readout_peak_median +
                self.pseudocount
        )

        self.normalized_cycle_data = self.cycle_data
        self.normalized_cycle_data = self.normalized_cycle_data.assign(
            readout_per_control =
                self.normalized_cycle_data.readout_peak_median /
                self.normalized_cycle_data.control_peak_median
        )

        self.sample_data = self.cycle_data.assign(
            cycle_nr = [
                0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
                1, 1, 1, 1, 1, 1, 1, 1, 1,
            ],
            sample_nr = [
                0, 0, 0, 0, 1, 1, 1, 1, 1, 1,
                0, 0, 0, 0, 1, 1, 1, 1, 1,
            ],
        )
        self.sample_data = self.sample_data.loc[
            self.sample_data.barcode == False
        ]
        self.sample_data = self.sample_data.drop(columns = "barcode")

        self.sample_data = self.sample_data.assign(
            readout_peak_z_score = [-1.,  1., -1.,  1., -1.,  1., -1.,  1.]
        )

        self.sample_data = self.sample_data[
            [
                "start_time",
                "end_time",
                "barcode_peak_median",
                "control_peak_median",
                "readout_peak_median",
                "readout_peak_z_score",
                "cycle_nr",
                "sample_nr",
            ]
        ]

        self.test_gen_map_content = (
            "9:CELLS\n10:SUBSTRATE\n11:FS\n12:FS\n"
            "13:Drug 1\n14:Drug 2\n15:Drug 3\n23:BCM\n24:BCM"
        )

        with tempfile.NamedTemporaryFile(mode = "w+t", suffix = ".txt") as \
            self.channel_file:

            self.channel_file.write(self.test_gen_map_content)
            self.channel_file.seek(0)
            self.channel_map = bd.ChannelMap(pl.Path(self.channel_file.name))

        self.plug_sequence = bd.PlugSequence(
            (
                bd.Sample(1, 2, "Drug 1 + Drug 2", [9, 10, 13, 14]),
                bd.Sample(1, 1, "Barcode", [11, 12, 23, 24]),
                bd.Sample(1, 2, "Drug 1 + Drug 3", [9, 10, 13, 15]),
                bd.Sample(1, 3, "End Cycle Barcode", [11, 12, 23, 24]),
            ),
            channel_map = self.channel_map,
        )

        self.labelled_sample_data = self.sample_data.assign(
            name = [
                "Drug 1 + Drug 2",
                "Drug 1 + Drug 2",
                "Drug 1 + Drug 3",
                "Drug 1 + Drug 3",
                "Drug 1 + Drug 2",
                "Drug 1 + Drug 2",
                "Drug 1 + Drug 3",
                "Drug 1 + Drug 3"
            ],
            compound_a = [
                "Drug 1",
                "Drug 1",
                "Drug 1",
                "Drug 1",
                "Drug 1",
                "Drug 1",
                "Drug 1",
                "Drug 1"
            ],
            compound_b = [
                "Drug 2",
                "Drug 2",
                "Drug 3",
                "Drug 3",
                "Drug 2",
                "Drug 2",
                "Drug 3",
                "Drug 3"
            ],
        )

        self.tmpdir = tempfile.mkdtemp()
        self.pmt_path = pl.Path(self.tmpdir, 'exp.txt')
        self.pmt_path.touch()
        self.config = config.PlugyConfig(
            input_dir = self.tmpdir,
            # workaround since there is only one simulated sample
            # readout_analysis_column is by default "readout_peak_z_score"
            # but z score calculation happens only for more than one sample
            # in sample_df
            readout_analysis_column = 'readout_peak_median'
        )


    @unittest.skip
    def test_plot_test_data(self):
        test_data_fig, test_data_ax = plt.subplots(1, 2, figsize = (40, 10))
        test_data_ax[0].plot(
            self.clean_data.time,
            self.clean_data.green,
            color = "green"
        )
        test_data_ax[0].plot(
            self.clean_data.time,
            self.clean_data.uv,
            color = "blue"
        )
        test_data_ax[0].plot(
            self.clean_data.time,
            self.clean_data.orange,
            color = "orange"
        )

        test_data_ax[1].plot(
            self.noisy_data.time,
            self.noisy_data.green,
            color = "green"
        )
        test_data_ax[1].plot(
            self.noisy_data.time,
            self.noisy_data.uv,
            color = "blue"
        )
        test_data_ax[1].plot(
            self.noisy_data.time,
            self.noisy_data.orange,
            color = "orange"
        )
        for i in range(2):
            test_data_ax[i].set_xlabel("Time [s]")
            test_data_ax[i].set_ylabel("PMT Output [V]")
        test_data_fig.tight_layout()
        test_data_fig.show()
        self.assertTrue(True)


    @unittest.skip
    def test_plot_detected_data(self):
        """
        Tests plotting of the plug data together with the pmt data
        """
        with unittest.mock.patch.object(
            target = pmt.PmtData,
            attribute = "read_txt",
            new = lambda _: self.noisy_data,
        ):

            # noinspection PyTypeChecker
            plug_data = plug.PlugData(
                pmt_data = pmt.PmtData(
                    input_file = pl.Path("MOCK"),
                    peak_min_distance = 0.03,
                    config = self.config,
                ),
                plug_sequence = None,
                channel_map = None,
                config = self.config,
            )

        plug_data_fig, plug_data_ax = plt.subplots(figsize = (40, 10))
        plug_data_ax = plug_data.plot_plug_pmt_data(axes = plug_data_ax)

        plug_data_fig.tight_layout()
        plug_data_fig.show()

        self.assertTrue(True)


    # noinspection DuplicatedCode
    def test_plug_detect_clean_data(self):
        """
        Tests detecting simple plugs from clean data
        """
        with unittest.mock.patch.object(
            target = pmt.PmtData,
            attribute = "read_txt",
            new = lambda _: self.clean_data,
        ):

            # noinspection PyTypeChecker
            plug_data = plug.PlugData(
                pmt_data = pmt.PmtData(
                    input_file = pl.Path("MOCK"),
                    peak_min_distance = 0.03,
                    config = self.config,
                ),
                has_controls = False,
                samples_per_cycle = 2,
                plug_sequence = None,
                channel_map = None,
                config = self.config,
            )

        pd_test.assert_frame_equal(
            self.cycle_data.round(),
            plug_data.plug_df[self.cycle_data.columns].round()
        )


    # noinspection DuplicatedCode
    def test_plug_detect_noisy_data(self):
        """
        Tests detecting plugs with a large amount of noise
        """
        with unittest.mock.patch.object(
            target = pmt.PmtData,
            attribute = "read_txt",
            new = lambda _: self.noisy_data
        ):

            # noinspection PyTypeChecker
            plug_data = plug.PlugData(
                pmt_data = pmt.PmtData(
                    input_file = pl.Path("MOCK"),
                    peak_min_distance = 0.03,
                    config = self.config,
                ),
                has_controls = False,
                samples_per_cycle = 2,
                plug_sequence = None,
                channel_map = None,
                config = self.config,
            )

        pd_test.assert_frame_equal(
            self.cycle_data.round(),
            plug_data.plug_df[self.cycle_data.columns].round()
        )


    # noinspection DuplicatedCode
    def test_plug_cycle_sample_calling(self):
        """
        Tests if cycles and sample numbers are properly detected
        """
        with unittest.mock.patch.object(
            target = pmt.PmtData,
            attribute = "read_txt",
            new = lambda _: self.noisy_data,
        ):

            # noinspection PyTypeChecker
            plug_data = plug.PlugData(
                pmt_data = pmt.PmtData(
                    input_file = pl.Path("MOCK"),
                    peak_min_distance = 0.03,
                    config = self.config,
                ),
                has_controls = False,
                samples_per_cycle = 2,
                plug_sequence = None,
                channel_map = None,
                min_end_cycle_barcodes = 3,
                n_bc_adjacent_discards = 0,
                config = self.config,
            )
            plug_data.detect_samples()

        pd_test.assert_frame_equal(
            self.sample_data.round(),
            plug_data.sample_df.iloc[:,0:8].round()
        )


    def test_plug_sample_labelling(self):
        """
        Tests if samples are properly labelled with the help of a
        bd.PlugSequence object
        """

        with unittest.mock.patch.object(
            target = pmt.PmtData,
            attribute = "read_txt",
            new = lambda _: self.noisy_data,
        ):

            # noinspection PyTypeChecker
            plug_data = plug.PlugData(
                pmt_data = pmt.PmtData(
                    input_file = pl.Path("MOCK"),
                    peak_min_distance = 0.03,
                    config = self.config,
                ),
                has_controls = False,
                plug_sequence = self.plug_sequence,
                channel_map = self.channel_map,
                min_end_cycle_barcodes = 3,
                n_bc_adjacent_discards = 0,
                config = self.config,
            )
            plug_data.detect_samples()

        pd_test.assert_frame_equal(
            self.labelled_sample_data.round(),
            plug_data.sample_df.round()
        )


    # noinspection DuplicatedCode
    def test_plug_detect_clean_data_cell_norm(self):
        """
        Tests detecting simple plugs from clean data
        """
        with unittest.mock.patch.object(
            target = pmt.PmtData,
            attribute = "read_txt",
            new = lambda _: self.clean_data,
        ):

            # noinspection PyTypeChecker
            plug_data = plug.PlugData(
                pmt_data = pmt.PmtData(
                    input_file = pl.Path("MOCK"),
                    peak_min_distance = 0.03,
                    config = self.config,
                ),
                has_controls = False,
                samples_per_cycle = 2,
                plug_sequence = None,
                channel_map = None,
                normalize_using_control=True,
                config = self.config,
            )

        print(self.normalized_cycle_data)
        print(plug_data.plug_df[self.normalized_cycle_data.columns])

        pd_test.assert_almost_equal(
            self.normalized_cycle_data,
            plug_data.plug_df[self.normalized_cycle_data.columns],
            atol = .01
        )


    # noinspection DuplicatedCode
    def test_plug_detect_noisy_data_cell_norm(self):
        """
        Tests detecting plugs with a large amount of noise
        """


        print(self.normalized_cycle_data)

        with unittest.mock.patch.object(
            target = pmt.PmtData,
            attribute = "read_txt",
            new = lambda _: self.noisy_data,
        ):

            # noinspection PyTypeChecker
            plug_data = plug.PlugData(
                pmt_data = pmt.PmtData(
                    input_file = pl.Path("MOCK"),
                    peak_min_distance = 0.03,
                    config = self.config,
                ),
                has_controls = False,
                samples_per_cycle = 2,
                plug_sequence = None,
                channel_map = None,
                normalize_using_control=True,
                config = self.config,
            )

        pd_test.assert_frame_equal(
            self.normalized_cycle_data.round(),
            plug_data.plug_df[self.normalized_cycle_data.columns].round()
        )


if __name__ == '__main__':

    unittest.main()



#!/usr/bin/env python
# -*- coding: utf-8 -*-

#
# This file is part of the `plugy` python module
#
# Copyright
# 2018-2021
# EMBL & Heidelberg University
#
# Author(s): Dénes Türei (turei.denes@gmail.com)
#            Nicolas Peschke
#            Olga Ivanova
#
# Distributed under the GPLv3 License.
# See accompanying file LICENSE.txt or copy at
#     http://www.gnu.org/licenses/gpl-3.0.html
#
# Webpage: https://github.com/saezlab/plugy
#


from .exp import PlugExperiment
from .data.config import PlugyConfig


#!/usr/bin/env python
# -*- coding: utf-8 -*-

#
# This file is part of the `plugy` python module
#
# Copyright
# 2018-2021
# EMBL & Heidelberg University
#
# Author(s): Dénes Türei (turei.denes@gmail.com)
#            Nicolas Peschke
#            Olga Ivanova
#
# Distributed under the GPLv3 License.
# See accompanying file LICENSE.txt or copy at
#     http://www.gnu.org/licenses/gpl-3.0.html
#
# Webpage: https://github.com/saezlab/plugy
#


import os
import unittest
import tempfile
import logging
import shutil

import pathlib as pl

from ..exp import PlugExperiment
from ..data.config import PlugyConfig

logging.basicConfig(
    level = logging.DEBUG,
    format = '%(asctime)s %(name)-12s %(levelname)-8s %(message)s',
    datefmt = '%d.%m.%y %H:%M:%S',
)


class TestPlugExperiment(unittest.TestCase):


    def setUp(self):

        self.tmpdir = tempfile.mkdtemp()


    def tearDown(self):

        shutil.rmtree(self.tmpdir)


    def test_check_input_file_names(self):
        """
        Tests if input file search is properly implemented.
        """

        fnames = (
            ('exp.txt', 'channels.tsv', 'sequence_879.csv'),
            ('Experiment_42.txt.gz', 'CHANNELS-xy.CSV', None),
            ('exp.xlsx', 'channel_layout.csv', 'Seq029.tsv'),
            ('fluorescence.txt', 'channel_layout.csv', 'Seq029.tsv'),
            ('fluor.TXT.gz', 'CHANNELS-xy.CSV', 'seq.tsv'),
        )

        attrs = ('pmt', 'channel', 'seq')

        for test_fnames in fnames:

            for fn in test_fnames:

                if fn is None:

                    continue

                path = pl.Path(self.tmpdir, fn)
                path.touch()

            if test_fnames[0].endswith('xlsx'):

                self.assertRaises(
                    FileNotFoundError,
                    lambda: PlugyConfig(input_dir = self.tmpdir)
                )

            else:

                cfg = PlugyConfig(input_dir = self.tmpdir)

                for fn, attr in zip(test_fnames, attrs):

                    attr = '%s_file' % attr
                    path = (
                        None
                            if fn is None else
                        pl.Path(self.tmpdir, fn)
                    )

                    self.assertEqual(getattr(cfg, attr), path)

            for f in os.listdir(self.tmpdir):

                os.remove(os.path.join(self.tmpdir, f))


if __name__ == '__main__':

    unittest.main()



#!/usr/bin/env python
# -*- coding: utf-8 -*-

#
# This file is part of the `plugy` python module
#
# Copyright
# 2018-2021
# EMBL & Heidelberg University
#
# Author(s): Dénes Türei (turei.denes@gmail.com)
#            Nicolas Peschke
#            Olga Ivanova
#
# Distributed under the GPLv3 License.
# See accompanying file LICENSE.txt or copy at
#     http://www.gnu.org/licenses/gpl-3.0.html
#
# Webpage: https://github.com/saezlab/plugy
#

import os
import time
import inspect
import subprocess as sp

import matplotlib.pyplot as plt
import seaborn as sns
import numpy as np


def add_git_hash_caption(fig: plt.Figure, offset: float = 0.8):
    """
    Adds a caption with the current git hash in a human readable form based
    on the most recent tag.
    If the repository contains changes that are not committed yet,
    the keyword "-dirty" is appended to the caption.

    :param fig: matplotlib.pyplot.Figure to add the caption to.
    :param offset: Determines how many millimeter the caption is offset
        from the bottom left corner of the figure.

    :return: None
    """

    sha = sp.run(
        ['git', 'describe', '--tags', '--long', '--dirty'],
        capture_output = True,
        text = True,
        cwd = os.path.dirname(os.path.realpath(__file__)),
    ).stdout.strip()
    rel_position = offset / (fig.get_size_inches() * 25.4)
    fig.text(
        rel_position[0],
        rel_position[1],
        f'Created on {time.ctime()} with plugy {sha}',
        fontsize = 'xx-small',
        fontweight = 'light',
    )


def plot_line(slope: float, intercept: float, axes: plt.Axes):
    """
    Plots a line using slope and intercept

    :param slope: Slope of the line
    :param intercept: Y-Axis intercept of the line
    :param axes: plt.Axes object to plot on
    :return: None
    """
    x_values = np.array(axes.get_xlim())
    y_values = intercept + slope * x_values
    axes.plot(x_values, y_values, '--', c = '#D22027')


def to_int(value):

    return int(float(value))


def to_tuple(value):

    return (
        value
            if isinstance(value, tuple) else
        ()
            if value is None else
        (value,)
    )


def to_set(value):
    """
    Makes sure the object `value` is a set, if it is a list converts
    it to set, otherwise it creates a single element set out of it.
    If `value` is None returns empty set.
    """

    if isinstance(value, set):

        return value

    elif value is None:

        return set()

    elif not isinstance(value, str) and hasattr(value, '__iter__'):

        return set(value)

    else:

        return {value}


def first(value):
    """
    Returns first element of an iterator or the value unchanged if it's not
    an iterator.
    """

    if hasattr(value, '__iter__'):

        try:

            value = next(value.__iter__())

        except StopIteration:

            value = None

    return value


def prettyfloat(n):

    return '%.02g' % n if isinstance(n, float) else str(n)


def dict_str(dct):

    if not isinstance(dct, dict):

        return str(dct)

    return ', '.join(
        '%s=%s' % (str(key), prettyfloat(val))
        for key, val in dct.items()
    )


def ntuple_str(nt, pretty_floats = True):

    return (
        ', '.join(
            '%s=%s' % (field, prettyfloat(value))
            for field, value in zip(nt._fields, nt)
        )
            if pretty_floats else
        nt.__repr__().split('(')[1][:-1]
    )


def matplotlib_331_fix():

    def get_extents(self, transform = None, **kwargs):

        Bbox = plt.matplotlib.transforms.Bbox
        Path = plt.matplotlib.path.Path

        if transform is not None:
            self = transform.transform_path(self)
        if self.codes is None:
            xys = self.vertices
        elif len(np.intersect1d(self.codes, [Path.CURVE3, Path.CURVE4])) == 0:
            xys = self.vertices[self.codes != Path.CLOSEPOLY]
        else:
            xys  =  []
            for curve, code in self.iter_bezier(**kwargs):
                # places where the derivative is zero can be extrema
                _, dzeros  =  curve.axis_aligned_extrema()
                # as can the ends of the curve
                xys.append(curve([0, *dzeros, 1]))
            xys  =  np.concatenate(xys)
        if len(xys):
            return Bbox([xys.min(axis = 0), xys.max(axis = 0)])
        else:
            return Bbox.null()


    if plt.matplotlib.__version__ == '3.3.1':

        plt.matplotlib.path.Path.get_extents = get_extents


def seaborn_violin_fix(
        box_color = None,
        midpoint_color = None,
        violin_border_width = None,
        box_linewidth = 1.5,
        **kwargs,
    ):
    """
    Draws miniature boxplots in the middle of the violins on a violinplot.
    Seaborn has hardcoded values for the colors of these boxplots, hence
    we need to provide a modified method to have any control over the
    aesthetics.
    """

    def add_boxes(ax, data, support, density, center, plotter, parent_env):

        def swap(*args):

            return reversed(args) if plotter.orient == 'v' else args


        q25, q50, q75 = np.percentile(data, [25, 50, 75])
        whisker_lim = 1.5 * (q75 - q25)
        h1 = np.min(data[data >= (q25 - whisker_lim)])
        h2 = np.max(data[data <= (q75 + whisker_lim)])

        box_color = parent_env['box_color'] or plotter.gray
        midpoint_color = parent_env['midpoint_color'] or 'white'

        # Draw a boxplot using lines and a point
        ax.plot(
            *swap([h1, h2], [center, center]),
            linewidth = parent_env['box_linewidth'],
            color = box_color,
        )
        ax.plot(
            *swap([q25, q75], [center, center]),
            linewidth = parent_env['box_linewidth'] * 3,
            color = box_color,
        )
        ax.scatter(
            *swap(q50, center),
            zorder = 3,
            color = midpoint_color,
            edgecolor = box_color,
            s = np.square(parent_env['box_linewidth'] * 2),
        )


    valid_args = {
        'x', 'y', 'hue', 'data', 'order', 'hue_order',
        'bw', 'cut', 'scale', 'scale_hue', 'gridsize',
        'width', 'inner', 'split', 'dodge', 'orient', 'linewidth',
        'color', 'palette', 'saturation',
    }

    ax = kwargs.pop('ax', None)
    violin_args = {
        k: v.default
        for k, v in inspect.signature(sns.violinplot).parameters.items()
        if v.default is not inspect.Parameter.empty
    }
    violin_args.update(kwargs)
    violin_args = dict(
        it for it in violin_args.items()
        if it[0] in valid_args
    )

    plotter = sns.categorical._ViolinPlotter(**violin_args)

    if ax is None:

        ax = plt.gca()

    plotter.plot(ax)

    for i, group_data in enumerate(plotter.plot_data):

        if plotter.plot_hues is None:

            support, density = plotter.support[i], plotter.density[i]
            violin_data = sns.utils.remove_na(group_data)
            add_boxes(ax, violin_data, support, density, i, plotter, locals())

        else:

            for j, hue_level in enumerate(plotter.hue_names):

                support, density = (
                    plotter.support[i][j],
                    plotter.density[i][j],
                )
                if support.size <= 1:

                    continue

                hue_mask = plotter.plot_hues[i] == hue_level
                violin_data = sns.utils.remove_na(group_data[hue_mask])

                add_boxes(
                    ax, violin_data, support, density,
                    i + (offsets[j] if plotter.split else 0), # noqa: F821
                    locals(),
                )

    # this is an example how to remove the violin borders
    # unfortunately seaborn doesn't provide more convenient API for that
    if violin_border_width is not None:

        _ = [
            ax.collections[i].set_linewidth(violin_border_width)
            for i in range(len(ax.collections))
        ]

    return ax


matplotlib_331_fix()